# -*- coding: utf-8 -*-

import os
import subprocess
import time

from unittest import (
    mock,
)

import pytest

from digiflow import (
    Blueprint,
    ResourceGenerator,
    DerivansManager,
    generate_structure,
    run_profiled,
    id_generator,
    ContainerDerivansManager,
    DEFAULT_DERIVANS_IMAGE,
)


DUMMY_METS = '<xml/>'


def test_generate_migration_start_structure_layout(tmp_path):
    """Create default structure using DEFAULT_STRUCTURE_DIR"""

    # act
    created = generate_structure(tmp_path)

    # assert
    assert 10 == len(created)
    for res in created:
        assert os.path.isfile(os.path.join(tmp_path, res))


def test_generate_kitodo2_metadata_layout(tmp_path):
    """Create structure like kitodo2"""

    # arrange
    gen_tifs = ResourceGenerator(
        '1064/images/100384256_media', Blueprint('.tif'))
    gen_jpgs = ResourceGenerator('100384256/jpg')

    # act
    created = generate_structure(tmp_path, [gen_tifs, gen_jpgs])

    # assert
    assert 20 == len(created)
    for res in created:
        assert os.path.isfile(os.path.join(tmp_path, res))


def test_generate_archivierung_layout(tmp_path):
    """Create structure like kitodo2"""

    # arrange
    gen_tifs = ResourceGenerator(
        '1003841856/metadata/images/1003841856_media',
        Blueprint('.tif'), number=383)
    gen_jpgs = ResourceGenerator('1003841856/content/jpg', number=383)
    gen_mark = ResourceGenerator('1003841856/zkw_open', blueprint=None)

    # act
    created = generate_structure(tmp_path, [gen_tifs, gen_jpgs, gen_mark])

    # assert
    assert len(created) >= 21
    for res in created:
        assert os.path.isfile(os.path.join(tmp_path, res))
    assert os.path.isfile(
        os.path.join(str(tmp_path), '1003841856', 'zkw_open'))


def test_generate_invalid_startdir():
    with pytest.raises(RuntimeError) as exc:
        generate_structure('foo/bar')

    assert 'Invalid start_dir \'foo/bar\' provided' in str(exc.value)


def test_generator_with_prefix_and_suffix():
    """
    Generate values with prefix and suffix
    """

    generator = id_generator(start=42, prefix='FOO_', suffix='_BAR')
    assert next(generator) == 'FOO_0043_BAR'
    assert next(generator) == 'FOO_0044_BAR'


def test_generator_with_mixed_previous_value():
    """
    Generate next value from previous value
    Guess prefix and where numerical part starts
    """

    generator = id_generator(previous_value="phys737438")
    assert next(generator) == 'phys737439'


def test_generator_with_numerical_start_and_padded_value():
    """
    Generate next value from start with padded to 4 digits
    """

    generator = id_generator(start=126, padded=4)
    assert next(generator) == '0127'


# pylint: disable=missing-function-docstring
@run_profiled
def my_func():
    time.sleep(.2)
    return 42


@run_profiled
def my_func_void():
    time.sleep(.2)


@run_profiled
def my_func_arg(rang):
    time.sleep(rang)
    return 27


@run_profiled
def my_func_args(aaa, bbb):
    time.sleep(.25)
    return aaa + bbb


# pylint: disable=missing-class-docstring
class MyClass:

    def __init__(self):
        self.val = 2

    @run_profiled
    def my_func_args(self, arg1, arg2):
        time.sleep(.25)
        return self.val + (arg1 * arg2)

    @staticmethod
    @run_profiled
    def my_func_static(arg1, arg2):
        time.sleep(.25)
        return arg1 * arg2


def test_utils_profile_method_no_arg():
    assert my_func() == (0.20, 'my_func', 42)


def test_utils_profile_arg():
    assert my_func_arg(0.75) == (0.75, 'my_func_arg', 27)


def test_utils_profile_noargs_noreturns():
    assert my_func_void() == (0.20, 'my_func_void', None)


def test_utils_profile_args_return():
    assert my_func_args(3, 5) == (0.25, 'my_func_args', 8)


def test_utils_profile_class_context():
    my_stuff = MyClass()
    assert my_stuff.my_func_args(3, 5) == (0.25, 'MyClass.my_func_args', 17)
    assert \
        MyClass.my_func_static(3, 5) == (0.25, 'MyClass.my_func_static', 15)


@mock.patch('docker.models.images.ImageCollection.pull')
def test_derivans_manager_with_container(mock_coll, tmp_path):
    """Check init phase of containerized
    Derivans execution context tries to
    pull required default container image"""

    # arrange
    test_project_root = tmp_path / 'migrationtest'
    test_project_root.mkdir()
    mets_file = os.path.join(str(test_project_root), 'mets_mods.xml')
    with open(mets_file, 'w', encoding='utf-8') as fh_mets:
        fh_mets.write(DUMMY_METS)
    dmanager = ContainerDerivansManager(
        mets_file,
    )

    # act
    dmanager.init()

    # assert
    assert mock_coll.call_count == 1
    _call_args = mock_coll.call_args.args
    assert len(_call_args) == 2
    _image_label = DEFAULT_DERIVANS_IMAGE.split(':', maxsplit=1)[0]
    assert _call_args[0] == f'{_image_label}'
    assert _call_args[1] == 'latest'


def test_derivans_manager_with_path_bin_dir(tmp_path):
    """
    Raise Exception because no valid Maven project
    Directory exists but required to build Derivans
    """

    # arrange
    test_project_root = tmp_path / 'migrationtest'
    the_derivans = test_project_root / 'digital-derivans' / 'target'
    test_project_root.mkdir()
    os.makedirs(str(the_derivans), exist_ok=True)
    test_project_bin = test_project_root / 'bin'
    test_project_bin.mkdir()
    mets_file = os.path.join(str(test_project_root), 'mets_mods.xml')
    with open(mets_file, 'w', encoding='utf-8') as fh_mets:
        fh_mets.write(DUMMY_METS)

    path_mvn_project = str(test_project_root / 'digital-derivans')
    dmanager = DerivansManager(
        mets_file, path_binary=str(test_project_bin),
        path_mvn_project=path_mvn_project)
    cwd = os.getcwd()

    # act
    with pytest.raises(subprocess.CalledProcessError) as err:
        dmanager.init()
    os.chdir(cwd)  # subprocess is changing the cwd, so lets go back

    # assert
    assert ("Command 'mvn clean package -DskipTests' "
            "returned non-zero exit status") in str(err.value)


def test_derivans_manager_none_path_binary(tmp_path):
    """Exception must be thrown if invalid
    None-path to local binary provided"""

    # arrange
    test_project_root = tmp_path / 'migrationtest'
    test_project_root.mkdir()
    mets_file = os.path.join(str(test_project_root), 'mets_mods.xml')
    with open(mets_file, 'w', encoding='utf-8') as fh_mets:
        fh_mets.write(DUMMY_METS)

    # act
    with pytest.raises(RuntimeError) as exc:
        DerivansManager(mets_file, path_binary=None)

    # assert
    assert 'invalid path_binary' in str(exc.value)


def _forward_derivans_call(*args):
    """take some rest and return alike
    the *real* call would do"""

    _delay = args[0]
    time.sleep(_delay)
    yield (_delay, 'execute', 'result')


@mock.patch('digiflow.DerivansManager._execute_derivans')
@mock.patch('digiflow.DerivansManager._identify_derivans_bin')
def test_derivans_start_set_exec(mock_check, mock_call, tmp_path):
    # arrange
    test_project_root = tmp_path / 'migrationtest'
    the_derivans = test_project_root / 'digital-derivans' / 'target'
    test_project_root.mkdir()
    os.makedirs(str(the_derivans), exist_ok=True)
    test_project_bin = test_project_root / 'bin'
    test_project_bin.mkdir()

    # create some dummy file
    mets_file = os.path.join(str(test_project_root), 'mets_mods.xml')
    with open(mets_file, 'w', encoding='utf-8') as fh_mets:
        fh_mets.write(DUMMY_METS)
    path_mvn_project = str(test_project_root / 'digital-derivans')
    dmanager = DerivansManager(
        mets_file, path_binary=str(test_project_bin),
        path_mvn_project=path_mvn_project)
    # here is the important detail
    dmanager.path_exec = '/usr/lib/jvm/java-11-openjdk-amd64/bin/java'
    mock_check.return_value = str(test_project_bin)
    mock_call.side_effect = _forward_derivans_call(0.1)

    # act
    dmanager.init()
    result = dmanager.start()

    # assert
    assert mock_call.call_count == 1
    assert result.command.startswith('/usr/lib/jvm/java-11-openjdk-amd64/bin/java -jar')


@mock.patch('digiflow.DerivansManager._execute_derivans')
@mock.patch('digiflow.DerivansManager._identify_derivans_bin')
def test_derivans_start_default(mock_check, mock_call, tmp_path):
    # arrange
    test_project_root = tmp_path / 'migrationtest'
    the_derivans = test_project_root / 'digital-derivans' / 'target'
    test_project_root.mkdir()
    os.makedirs(str(the_derivans), exist_ok=True)
    test_project_bin = test_project_root / 'bin'
    test_project_bin.mkdir()

    # create some dummy file
    mets_file = os.path.join(str(test_project_root), 'mets_mods.xml')
    with open(mets_file, 'w', encoding='utf-8') as fh_mets:
        fh_mets.write(DUMMY_METS)
    path_mvn_project = str(test_project_root / 'digital-derivans')
    dmanager = DerivansManager(
        mets_file, path_binary=str(test_project_bin),
        path_mvn_project=path_mvn_project)
    mock_check.return_value = str(test_project_bin)
    mock_call.side_effect = _forward_derivans_call(0.1)

    # act
    dmanager.init()
    result = dmanager.start()

    # assert
    assert mock_call.call_count == 1
    assert result.command.startswith('java -jar')


@mock.patch('digiflow.DerivansManager._execute_derivans')
@mock.patch('digiflow.DerivansManager._identify_derivans_bin')
def test_derivans_start_with_additional_config(mock_check, mock_call, tmp_path):
    """Prevent Regression and catch bug: 
        [undefined]TypeError: bad operand type for unary +: 'str'
    """

    # arrange
    test_project_root = tmp_path / 'migrationtest'
    the_derivans = test_project_root / 'digital-derivans' / 'target'
    test_project_root.mkdir()
    os.makedirs(str(the_derivans), exist_ok=True)
    test_project_bin = test_project_root / 'bin'
    test_project_bin.mkdir()

    # create some dummy file
    mets_file = os.path.join(str(test_project_root), 'mets_mods.xml')
    with open(mets_file, 'w', encoding='utf-8') as fh_mets:
        fh_mets.write(DUMMY_METS)
    path_mvn_project = str(test_project_root / 'digital-derivans')
    dmanager = DerivansManager(
        mets_file, path_binary=str(test_project_bin),
        path_mvn_project=path_mvn_project)
    mock_check.return_value = str(test_project_bin)
    dmanager.path_configuration = '/path/to/derivans.ini'
    mock_call.side_effect = _forward_derivans_call(0.1)

    # act
    dmanager.init()
    result = dmanager.start()

    # assert
    assert mock_call.call_count == 1
    assert result.command.startswith('java -jar')
    assert result.command.endswith('/path/to/derivans.ini')


@mock.patch('digiflow.DerivansManager._execute_derivans')
@mock.patch('digiflow.DerivansManager._identify_derivans_bin')
def test_derivans_start_with_java_xargs(mock_check, mock_call, tmp_path):
    """Prevent Regression and catch bug: 
        X11 connection rejected because of wrong authentication.
Exception in thread "main" java.awt.AWTError: Can't connect to X11 window server using 'localhost:11.0' as the value of the DISPLAY variable.
        at java.desktop/sun.awt.X11GraphicsEnvironment.initDisplay(Native Method)
        at java.desktop/sun.awt.X11GraphicsEnvironment$1.run(X11GraphicsEnvironment.java:102)
        at java.base/java.security.AccessController.doPrivileged(Native Method)
        at java.desktop/sun.awt.X11GraphicsEnvironment.<clinit>(X11GraphicsEnvironment.java:61)
        at java.base/java.lang.Class.forName0(Native Method)
        at java.base/java.lang.Class.forName(Class.java:315)
        at java.desktop/java.awt.GraphicsEnvironment$LocalGE.createGE(GraphicsEnvironment.java:101)
        at java.desktop/java.awt.GraphicsEnvironment$LocalGE.<clinit>(GraphicsEnvironment.java:83)
        at java.desktop/java.awt.GraphicsEnvironment.getLocalGraphicsEnvironment(GraphicsEnvironment.java:129)
        at de.ulb.digital.derivans.derivate.FontHandler.forGraphics(FontHandler.java:37)

        solved by passing "-Djava.awt.headless=true" to java call
    """

    # arrange
    test_project_root = tmp_path / 'migrationtest'
    the_derivans = test_project_root / 'digital-derivans' / 'target'
    test_project_root.mkdir()
    os.makedirs(str(the_derivans), exist_ok=True)
    test_project_bin = test_project_root / 'bin'
    test_project_bin.mkdir()

    # create some dummy file
    mets_file = os.path.join(str(test_project_root), 'mets_mods.xml')
    with open(mets_file, 'w', encoding='utf-8') as fh_mets:
        fh_mets.write(DUMMY_METS)
    path_mvn_project = str(test_project_root / 'digital-derivans')
    dmanager = DerivansManager(
        mets_file, path_binary=str(test_project_bin),
        path_mvn_project=path_mvn_project)
    mock_check.return_value = str(test_project_bin)
    dmanager.path_configuration = '/path/to/derivans.ini'
    dmanager.xargs = '-Djava.awt.headless=true'
    mock_call.side_effect = _forward_derivans_call(0.1)

    # act
    dmanager.init()
    result = dmanager.start()

    # assert
    assert mock_call.call_count == 1
    assert ' -Djava.awt.headless=true ' in result.command



# -*- coding: utf-8 -*-
from __future__ import annotations

import itertools
import os
import pathlib
import platform
import re
import shutil
import subprocess
import time
from abc import abstractmethod, ABC
from dataclasses import dataclass

from pathlib import Path

from typing import (
    List,
    Tuple,
    Union,
    Final, Callable, Any, TypeVar, Generic,
)

from PIL import (
    Image
)

import numpy as np

from docker import (
    DockerClient,
    from_env
)
from docker.models.containers import Container
from docker.models.resource import Model
from docker.types import Mount

# default sub dir for structure creation
DEFAULT_STRUCTURE_DIR = 'MAX'

# derivates
DEFAULT_DERIVANS_IMAGE = "ghcr.io/ulb-sachsen-anhalt/digital-derivans:latest"
DEFAULT_DERIVANS_TIMEOUT = 10800
DERIVANS_LABEL = 'derivans'
DERIVANS_CNT_DATA_DIR: Final[str] = "/data_mets"
DERIVANS_CNT_CONF_DIR: Final[str] = "/data_cfg"


def id_generator(
        start=0, prefix=None, suffix=None, previous_value=None, padded=4):
    """
    Generate Numbers with schema [<prefix>]<4digits>[<suffix>]
    with default padded zeros to match 4-digit-numbers
    or
    Try to guess prefix from given previous_value
    """

    _prefix = None
    _start = None
    if previous_value is not None:
        i = 0
        while i < len(previous_value):
            if not _prefix and previous_value[i].isnumeric():
                _prefix = previous_value[:i]
                _start = previous_value[i:]
            i += 1

    if _prefix and not prefix:
        prefix = _prefix
    if _start and not start:
        start = int(_start)

    for num in range(start + 1, 99999999):
        number = str(num).zfill(padded)
        if prefix:
            number = "{}{}".format(prefix, number)
        if suffix:
            number = "{}{}".format(number, suffix)
        yield number


def create_grayscale_data(dim):
    arr_floats = np.random.rand(dim[1], dim[0]) * 255
    arr_ints = arr_floats.astype(np.uint8)
    return arr_ints


def create_sample_text():
    return 'some foo text'


class Blueprint():

    def __init__(self,
                 ext='.tif',
                 batch=True,
                 content_func=create_grayscale_data,
                 **kwargs):
        self.ext = ext
        self.batch = batch
        self.content_func = content_func
        self.dimension = None
        self.header = None
        if 'header' in kwargs:
            self.header = kwargs['header']
        self.dimension = kwargs['dimension'] if 'dimension' in kwargs else (
            200, 300)

    def write(self, path):
        if self.ext in ('.tif', '.jpg'):
            self.create_image(path, params=self.header)

    def create_image(self, path_image, params=None):
        """
        Create artificial greyscale images
        """

        data = self.content_func(self.dimension)
        im = Image.fromarray(data)
        store_format = 'TIFF' if self.ext == '.tif' else 'JPEG'
        im.save(path_image, format=store_format, params=params)
        return path_image


class ResourceGenerator:

    def __init__(self, path, blueprint=Blueprint('.jpg'), number=10):
        self.path = path
        self.blueprint = blueprint
        self.content_func = create_sample_text
        self.number = number
        self.start_dir = None

    def get_batch(self, start=0, padded=8):
        paths = self._get_paths(start, self.number, padded)
        aggregated_paths = []
        for path in paths:
            path_dir = self._make_dirs(path)
            path_file = os.path.basename(path)
            real_path = os.path.join(str(path_dir), str(path_file))
            self.blueprint.write(real_path)
            aggregated_paths.append(real_path)
        return aggregated_paths

    def get(self):
        path_dir = self._make_dirs(self.path)
        path_file = os.path.basename(self.path)
        real_path = os.path.join(str(path_dir), str(path_file))
        self.write_single(real_path)
        return real_path

    def write_single(self, path):
        with open(path, 'w') as fhd:
            fhd.write(self.content_func())

    def _make_dirs(self, path):
        the_dir = pathlib.Path(os.path.dirname(path))
        path_dir = os.path.join(str(self.start_dir), str(the_dir))
        os.makedirs(path_dir, exist_ok=True)
        return path_dir

    def _get_paths(self, start, number, padded=8, previous_value=None):
        self.name_generator = id_generator(
            start, previous_value=previous_value, padded=padded)
        names = list(itertools.islice(self.name_generator, number))
        return [os.path.join(self.path, n + self.blueprint.ext) for n in names]


def generate_structure(
        start_dir,
        gens=None,
        number=10
) -> List[str]:

    if gens is None:
        gens = [ResourceGenerator(DEFAULT_STRUCTURE_DIR)]

    """
    Generate test data layouts using a list of Generators for
    * Kitodo2 metadata        (<id>/images/<title>_media/*.tif)
    * Kitodo2 export          (<title>/jpeg/*.jpg)
    * Archivierung            (<title>/content/ ,
                               <title>/metadata/images/<title>_media/*.tif)
    * Migration OAI Workspace (MAX/*.jpg)
    * SAF creation Workspace  (<several-subdirs>)
    """

    generated = []

    if not os.path.isdir(str(start_dir)):
        raise RuntimeError(
            "Invalid start_dir '{}' provided!".format(start_dir))

    for gen in gens:
        gen.start_dir = start_dir
        if gen.blueprint and gen.blueprint.batch:
            resouces = gen.get_batch()
            for res in resouces:
                resource_path = res
                generated.append(resource_path)
        else:
            generated.append(gen.get())

    return generated


_T = TypeVar('_T')
_FuncWrapperResult = Tuple[float, str, _T]
RunProfiledResult = Callable[[], _FuncWrapperResult[_T]]


def run_profiled(func: Callable) -> RunProfiledResult:
    """
    Decorator to profile method execution time
    in seconds as float with 2 decimal digits
    """

    def _get_func_name(_func: Callable) -> str:
        _label = str(_func)
        match = re.match(r'.*function ([\w.]+) *', _label)
        if match:
            return match.group(1)
        else:
            # this recognizes even MagicMock-Objects
            match = re.match(r".*name='([a-z_]+)'.*", _label)
            if match:
                return match.group(1)
        return 'func'

    def func_wrapper(*args) -> _FuncWrapperResult[_T]:
        result = None
        start: float = time.perf_counter()
        if args:
            result = func(*args)
        else:
            result = func()
        delta: float = time.perf_counter() - start

        # name of callee
        label: str = _get_func_name(func)
        return round(delta, 2), label, result

    return func_wrapper


@run_profiled
def run_command(cmd, timeout) -> subprocess.CompletedProcess:
    """Forward command with given timeout
    """
    return subprocess.run(
        cmd,
        shell=True,
        check=True,
        capture_output=True,
        encoding='UTF-8',
        timeout=timeout
    )


@dataclass(frozen=True)
class DerivansResult:
    command: str
    duration: float
    result: str | ContainerProcResult | None
    label: str = None


@dataclass(frozen=True)
class ContainerProcResult:
    exit_code: int
    logs: str


class BaseDerivansManager(ABC):
    """Manage Derivans component calls
    from within python modules
    """

    @staticmethod
    def create(
            path_mets_file: str,
            container_image_name: str = None,
            path_binary: str = None,
            path_mvn_project: str = None,
            path_configuration: str = None,
    ) -> BaseDerivansManager:
        """Create actual DerivansManager instance
        depending on provided parameters"""

        if container_image_name is not None:
            return ContainerDerivansManager(
                path_mets_file=path_mets_file,
                container_image=container_image_name,
                path_configuration=path_configuration,
            )
        return DerivansManager(
            path_mets_file=path_mets_file,
            path_binary=path_binary,
            path_mvn_project=path_mvn_project,
            path_configuration=path_configuration,
        )

    def __init__(
            self,
            path_mets_file,
            path_configuration=None,
    ):
        if path_configuration and not Path(str(path_configuration)).is_file():
            raise RuntimeError(f"[DerivansManager] config missing: {path_configuration}!")
        self.path_mets_file = path_mets_file
        self.path_configuration = path_configuration

    @abstractmethod
    def init(self) -> None:
        """Setup application"""

    @abstractmethod
    def start(self) -> DerivansResult:
        """Issue actual derivans instance
        and communicate outcome"""


class DerivansManager(BaseDerivansManager):
    """Local Derivans instance.
    Requires at least recent Java at
    runtime, additionally Maven if
    Derivans must be built at init stage"""

    @property
    def timeout(self):
        """Timeout for derivans workflows"""
        return self._timeout

    @timeout.setter
    def timeout(self, timeout):
        self._timeout = timeout

    @property
    def label(self):
        return self._label

    @label.setter
    def label(self, label):
        self._label = label

    def __init__(
            self,
            path_mets_file: str,
            path_binary: str,
            path_mvn_project: str = None,
            path_configuration: str = None,
    ):
        if path_binary is None or \
                not (Path(path_binary).is_dir() or Path(path_binary).is_file()):
            raise RuntimeError(f"[DerivansManager] invalid path_binary: {path_binary}!")
        if path_mvn_project is not None \
                and not Path(str(path_mvn_project)).is_dir():
            raise RuntimeError(f"[DerivansManager] invalid path_mvn_project: {path_mvn_project}!")
        super().__init__(
            path_mets_file=path_mets_file,
            path_configuration=path_configuration,
        )
        self.path_binary = path_binary
        self.path_mvn_project = path_mvn_project
        self._timeout = DEFAULT_DERIVANS_TIMEOUT
        self._label = DERIVANS_LABEL
        self.path_exec = None
        self.xargs = ''

    def init(self):
        _path_derivans = self.path_binary

        if Path(self.path_binary).is_dir():
            self.path_binary = self._identify_derivans_bin()

        if self.path_binary is None:
            self._recreate_app(_path_derivans)

        # fallback to default 'java' if no need to worry about
        if not self.path_exec:
            self.path_exec = 'java'

    def start(self) -> DerivansResult:
        """Create Derivates with provided configuration

        * step into derivans root dir first
        * respect actual operation system due executable path
          wich might contain spaces on windows
        * execute
        * return to previous directory

        """
        derivans_root = os.path.dirname(self.path_binary)
        prev_dir = os.path.abspath(os.curdir)
        os.chdir(derivans_root)
        path_exec = self.path_exec
        if platform.system() not in ['Linux']:
            path_exec = f'"{path_exec}"'
        if self.xargs and not self.xargs.startswith(' '):
            self.xargs = ' ' + self.xargs
        cmd = f'{path_exec}{self.xargs} -jar {self.path_binary} {self.path_mets_file}'
        if self.path_configuration:
            cmd += f' -c {self.path_configuration}'
        # disable pylint due it is not able to recognize
        # output being created by decorator
        time_duration, label, result = self._execute_derivans(cmd)  # pylint: disable=unpacking-non-sequence
        os.chdir(prev_dir)
        return DerivansResult(
            command=cmd,
            result=result,
            duration=time_duration,
            label=label,
        )

    def _identify_derivans_bin(self, the_dir=None):
        if not the_dir:
            the_dir = self.path_binary
        all_files = [f for f in os.listdir(the_dir) if f.endswith('.jar')]
        derivantis = sorted([f for f in all_files if self._label in str(f)])
        if len(derivantis) > 0:
            return os.path.join(the_dir, derivantis[0])
        return None

    def _recreate_app(self, target_dir):
        dir_derivans = self.path_mvn_project
        if not dir_derivans:
            raise RuntimeError("Derivans project dir unset!")
        if not os.path.isdir(dir_derivans):
            raise RuntimeError(f"Invalid derivans project dir: '{dir_derivans}'!")

        derivans_build_dir = os.path.join(dir_derivans, 'target')
        the_derivans = None
        if os.path.exists(derivans_build_dir):
            the_derivans = self._identify_derivans_bin(derivans_build_dir)

        # if derivans app not build yet, then ...
        if not the_derivans:
            os.chdir(dir_derivans)
            compl_proc = subprocess.run(
                'mvn clean package -DskipTests',
                shell=True,
                check=True,
                timeout=600)
            if compl_proc.returncode != 0:
                raise RuntimeError("Cant build app '%s' in '%s'!")
            the_derivans = self._identify_derivans_bin(derivans_build_dir)

        # copy new built jar to target directory
        if not os.path.exists(target_dir):
            os.mkdir(target_dir)
        shutil.copy(the_derivans, target_dir)
        self.path_binary = os.path.join(
            target_dir, os.path.basename(the_derivans))

    def _execute_derivans(self, command) -> _FuncWrapperResult:
        return run_command(command, self.timeout)


class ContainerDerivansManager(BaseDerivansManager):
    """Containered Derivans instance.
    Required local container runtime.
    """

    def __init__(
            self,
            path_mets_file: str,
            container_image: str = DEFAULT_DERIVANS_IMAGE,
            path_configuration: str = None,
    ):
        super().__init__(
            path_mets_file=path_mets_file,
            path_configuration=path_configuration
        )
        self._container_image: str = container_image
        self._client: DockerClient = from_env()

    def init(self) -> None:
        repo, tag = self._container_image.split(':')
        self._client.images.pull(repo, tag)

    def start(self) -> DerivansResult:
        mounts: List[Mount] = []
        command: List[str] = []
        mets_path: Path = Path(self.path_mets_file).absolute()
        if mets_path.is_dir():
            command.append(DERIVANS_CNT_DATA_DIR)
            mounts.append(Mount(source=str(mets_path), target=DERIVANS_CNT_DATA_DIR, type='bind'))
        if mets_path.is_file():
            mets_file_name: str = mets_path.name
            mets_dir: str = str(mets_path.parent.absolute())
            target_mets_file: str = str(Path(DERIVANS_CNT_DATA_DIR).joinpath(mets_file_name))
            command.append(target_mets_file)
            mounts.append(Mount(source=mets_dir, target=DERIVANS_CNT_DATA_DIR, type='bind'))
        if self.path_configuration is not None:
            config_path: Union[Path, None] = Path(self.path_configuration)
            if config_path.exists() and config_path.is_file():
                config_file_name: str = config_path.name
                config_dir: str = str(config_path.parent.absolute())
                target_config_file: str = str(Path(DERIVANS_CNT_CONF_DIR).joinpath(config_file_name))
                mounts.append(Mount(source=config_dir, target=DERIVANS_CNT_CONF_DIR, type='bind'))
                command.append('-c')
                command.append(target_config_file)

        start_time: float = time.perf_counter()
        container: Container = self._client.containers.run(
            image=self._container_image,
            command=command,
            user=os.getuid(),
            mounts=mounts,
            detach=True
        )
        exit_code: int = container.wait()['StatusCode']
        logs: str = container.logs().decode('utf-8')
        container.remove()

        full_command_equivalent: List[str] = ['docker run -rm']
        for mount in mounts:
            full_command_equivalent.append(
                f"--mount type={mount['Type']},source={mount['Source']},target={mount['Target']}"
            )
        full_command_equivalent.append(self._container_image)
        full_command_equivalent.append(" ".join(command))
        dur: float = time.perf_counter() - start_time
        return DerivansResult(
            command=" ".join(full_command_equivalent),
            result=ContainerProcResult(exit_code=exit_code, logs=logs),
            duration=dur,
        )



# -*- coding: utf-8 -*-

import csv
import os
import shutil
import time
import smtplib
from email.mime.text import (
    MIMEText
)

from collections import (
    OrderedDict
)
from pathlib import (
    Path
)

import requests
from lxml import etree as ET

from .digiflow_metadata import (
    XMLNS,
    write_xml_file,
    MetsReader
)


####
#
# OAIRecord Specification (based on Migration)
#
F_IDENTIFIER = 'IDENTIFIER'
F_SPEC = 'SETSPEC'
F_DATESTAMP = 'CREATED'
F_STATE_INFO = 'INFO'
F_STATE = 'STATE'
F_STATE_TS = 'STATE_TIME'
HEADER_MIGRATION = [F_IDENTIFIER, F_SPEC, F_DATESTAMP,
                    F_STATE_INFO, F_STATE, F_STATE_TS]
COMMENT_MARK = '#'
UNSET = 'n.a.'
RECORD_STATE_UNSET = UNSET
RECORD_STATE_MASK_FRAME = 'other_load'
SETSPEC_SPLITTER = '##'
STATETIME_FORMAT = '%Y-%m-%d_%H:%M:%S'
STATETIME_FORMAT_ALT = '%Y-%m-%dT%H:%M:%SZ'


def post_oai_extract_metsdata(xml_tree):
    """Extract METS as new root from OAI envelope"""

    namespace = xml_tree.xpath('namespace-uri(.)')
    if namespace == 'http://www.loc.gov/METS/':
        return xml_tree

    if namespace == 'http://www.openarchives.org/OAI/2.0/':
        mets_root_el = xml_tree.find('.//mets:mets', XMLNS)
        if mets_root_el is not None:
            return ET.ElementTree(mets_root_el).getroot()
    return None


def post_oai_extract_mets(the_self, the_data):
    """Just extract METS from OAI body"""

    xml_root = ET.fromstring(the_data)
    mets_tree = post_oai_extract_metsdata(xml_root)
    write_xml_file(mets_tree, the_self.path_mets, preamble=None)


def post_oai_store_ocr(path_local, the_data):
    """
    Store OCR XML as it is
    Explicite encoding is required with OCR-strings but not
    for byte objects like from semantics fulltext responses
    """

    if isinstance(the_data, str):
        the_data = the_data.encode('utf-8')
    xml_root = ET.fromstring(the_data)
    write_xml_file(xml_root, path_local, preamble=None)


def get_enclosed(tokens_str:str, mark_end='}', mark_start='{', func_find='rfind') -> str:
    """
    Search dict-like enclosed entry in string
    from end (rfind) or start (find)
    
    If no match, return empty string ''
    """
    if mark_end in tokens_str and mark_start in tokens_str:
        _offset_right_end = tokens_str.__getattribute__(func_find)(mark_end)
        _offset_right_start = tokens_str[:_offset_right_end].__getattribute__(func_find)(mark_start)
        _the_enclosed = tokens_str[_offset_right_start:(_offset_right_end+1)]
        return _the_enclosed
    return ''


class OAIRecord:
    """
    OAIRecord based on valid URN-Identifier with optional set specification data
    based on http://www.openarchives.org/OAI/2.0/guidelines-oai-identifier.htm

    Examples:

    * oai:digital.bibliothek.uni-halle.de/hd:10595
    * oai:digitale.bibliothek.uni-halle.de/vd18:9427342
    * oai:digitale.bibliothek.uni-halle.de/zd:9633001
    * oai:opendata.uni-halle.de:1981185920/34265
    * oai:menadoc.bibliothek.uni-halle.de/menadoc:20586
    * oai:dev.opendata.uni-halle.de:123456789/27949

    """

    def __init__(self, urn):
        self.__urn = urn
        self.__local_ident = None
        self.set = UNSET
        self.date_stamp = UNSET
        self.info = UNSET
        self.state = UNSET
        self.state_datetime = UNSET

    @property
    def local_identifier(self):
        """
        source identifier for local usage
        as sub-directory in local storages
        * remove any 'oai:<host>...' related urn prefix
        * replace possible '/' in handle urns
        """
        if not self.__local_ident:
            _local_ident = self.__urn
            if ':' in _local_ident:
                _splits = self.__urn.split(':')
                _local_ident = _splits[-1]
            if '/' in _local_ident:
                _local_ident = _local_ident.replace('/', '_')
            self.__local_ident = _local_ident
        return self.__local_ident

    @property
    def identifier(self):
        return self.__urn

    def __str__(self) -> str:
        return "{}\t{}\t{}\t{}\t{}\t{}".format(self.__urn, self.set,
                                               self.date_stamp, self.info, self.state, self.state_datetime)


def transform_statelist2oairecord(row):
    """Transform record from state list to Migration OAIRecord
    Sets _at least_ an identifier, optional
    * setSpec   record set(s) this record belongs to
    * datestamp time of publication or modification of record
    """

    record = OAIRecord(row[F_IDENTIFIER])
    if F_SPEC in row and str(row[F_SPEC]).strip():
        record.set = row[F_SPEC]
    if F_DATESTAMP in row and str(row[F_DATESTAMP]).strip():
        record.date_stamp = row[F_DATESTAMP]
    return record


def transform_to_record(row):
    """Serialize data row into OAIRecord
    with all attributes filled"""

    record = OAIRecord(row[F_IDENTIFIER])
    record.set = row[F_SPEC]
    record.date_stamp = row[F_DATESTAMP]
    record.info = row[F_STATE_INFO]
    record.state = row[F_STATE]
    record.state_datetime = row[F_STATE_TS]
    return record


class OAILoadException(Exception):
    """Load of OAI Data failed"""

    def __init__(self, msg):
        self.message = msg
        super().__init__(self.message)

class OAILoadServerError(OAILoadException):
    """Loading of Record via OAI failed due
    response status_code indicating Server Error"""


class OAILoadClientError(OAILoadException):
    """Loading of Record via OAI failed due 
    response status_code indicating Client Error"""


class OAILoadContentException(OAILoadException):
    """Loading of Record via OAI failed due unexpected 
    returned content, indicating missing record data 
    or even missing complete record"""


class OAIRecordHandlerException(Exception):
    """Mark Exception during procesing of OAI Record Lists"""

    def __init__(self, msg):
        self.message = msg
        super().__init__(self.message)


class OAIRecordCriteria:
    """
    Criteria to pick OAIRecords from a list
    """

    def matched(self, _: OrderedDict) -> bool:
        """Determine, whether given OAI-Record matched criteria"""


class OAIRecordCriteriaIdentifier(OAIRecordCriteria):

    def __init__(self, ident):
        self.ident = ident

    def matched(self, record: OrderedDict) -> bool:
        rec_id = record[F_IDENTIFIER]
        # maybe must deal shortened identifier (like legacy id)
        if 'oai' not in self.ident or ':' not in self.ident:
            rec_id = rec_id.split(':')[-1]
        return self.ident == rec_id


class OAIRecordCriteriaState(OAIRecordCriteria):

    def __init__(self, state):
        self.state = state

    def matched(self, record: OrderedDict) -> bool:
        record_state = record[F_STATE]
        return self.state == record_state


class OAIRecordCriteriaDatetime(OAIRecordCriteria):
    """
    Use datetime data to match OAIRecords.\n
    Match field 'STATE_TIME' (or 'CREATED')\n
    Default datetime pattern:
        'dt_pattern' = '%Y-%m-%d_%H:%M:%S' (fits 'STATE_TIME'),\n
        can be overwritten ('CREATED' uses commonly: '%Y-%m-%dT%H:%M:%SZ').
    """

    def __init__(self, **kwargs):
        self.dt_pattern = STATETIME_FORMAT
        self.field = F_STATE_TS
        self.dt_from = None
        self.dt_to = None
        # *ORDER MATTERS*
        if 'dt_field' in kwargs.keys():
            self.field = kwargs['dt_field']
        if 'dt_format' in kwargs.keys():
            self.dt_pattern = kwargs['dt_format']
        if 'dt_from' in kwargs.keys():
            self.dt_from = time.strptime(kwargs['dt_from'], self.dt_pattern)
        if 'dt_to' in kwargs.keys():
            self.dt_to = time.strptime(kwargs['dt_to'], self.dt_pattern)

    def matched(self, record: OrderedDict) -> bool:
        if self.field not in record:
            raise RuntimeError("Field {} not in {}".format(self.field, record))
        record_state_ts = record[self.field]
        if record_state_ts == RECORD_STATE_UNSET:
            return False
        record_ts = time.strptime(record_state_ts, self.dt_pattern)
        if self.dt_from and not self.dt_to:
            return record_ts >= self.dt_from
        elif self.dt_from and self.dt_to:
            return (record_ts >= self.dt_from) and (record_ts < self.dt_to)
        elif not self.dt_from and self.dt_to:
            return record_ts < self.dt_to
        return False


class OAIRecordCriteriaText(OAIRecordCriteria):

    def __init__(self, text, field=F_STATE_INFO) -> None:
        super().__init__()
        self.text = text
        self.field = field

    def matched(self, record: OrderedDict) -> bool:
        return self.text in record[self.field]


class OAIRecordHandler:
    """
    Process loading and storing of dataset records
    which contain information about the state of
    a single record
    * records has *at least* 3 fields:
    => first field/column identifies record
    => pre-last sets state
    => last field sets timestamp for state
    """

    def __init__(self, data_path, data_fields=None,
                 ident_field=0,
                 mark_open=RECORD_STATE_UNSET, mark_lock='busy',
                 transform_func=transform_statelist2oairecord):
        self.data_path = str(data_path)
        self.mark = {'open': mark_open, 'lock': mark_lock}
        self.position = None
        self.header = None
        self.transform_func = transform_func
        self._raw_lines = []
        # read raw lines
        with open(self.data_path, encoding='utf-8') as tmp:
            self._raw_lines = tmp.readlines()
        # pick rows containing *real* data
        # skip empty ones and comment lines
        data_lines = [line
                      for line in self._raw_lines
                      if self._is_data_row(line)]
        # check data format if possible
        self._restore_header(data_lines[0])
        if data_fields:
            self._validate_header(data_fields)
            self.header = data_fields
        # inspect data integrity
        self.ident_field = self.header[ident_field]
        self.state_field = self.header[-2]
        self.state_ts_field = self.header[-1]
        # build data
        self.index = {}
        self.data = []
        self.record_prefix = None
        self._build_data()

    @property
    def total_len(self):
        return len(self.data)

    def _build_data(self):
        """Transform raw_lines into meaningful data and build index
        for faster access time.

        with 0 = index of _raw_line_data
             1 = index of data dictionary
        """
        for i, _raw_row in enumerate(self._raw_lines):
            if self._is_data_row(_raw_row) and \
                    not self._is_header_row(_raw_row):
                _data = self._to_dict(_raw_row)
                _data_idx = len(self.data)
                self.data.append(_data)
                the_ident = _data[self.ident_field]
                self.index[the_ident] = (i, _data_idx)

    def _to_dict(self, row_as_str):
        """
        split row by tabulator
        """

        splits = row_as_str.strip().split('\t')
        return OrderedDict(zip(self.header, splits))

    @staticmethod
    def _to_str_nl(dict_row):
        return '\t'.join(dict_row.values()) + '\n'

    @staticmethod
    def _is_data_row(row):
        if row:
            row_str = row.strip()
            not_empty = len(row_str) > 0
            not_comment = not row_str.startswith(COMMENT_MARK)
            return not_empty and not_comment
        return False

    def _is_header_row(self, row_str):
        if self.header:
            return row_str.startswith(self.header[0])
        return False

    def _restore_header(self, first_line):
        _header = [h.strip() for h in first_line.split('\t')]
        if not _header:
            _header = HEADER_MIGRATION
        self.header = _header

    def _validate_header(self, data_fields):
        """validate both occurence and order"""
        if self.header != data_fields:
            msg = "invalid fields: '{}', expect: '{}'".format(
                self.header, data_fields)
            raise RuntimeError(msg)

    def next_record(self, state=None):
        """
        Get *NEXT* single OAIRecord _from scratch_ with
        given state if any exist, None otherwise
        """

        if not state:
            state = self.mark['open']
        for i, row in enumerate(self.data):
            if state == row[self.state_field]:
                self.position = "{:04d}/{:04d}".format((i + 1), self.total_len)
                return self.transform_func(row)

    def get(self, identifier, exact_match=True):
        """Read data for first OAIRecord with
        given identifier *without* changing state

        Args:
            identifier (string): record URN
            exact_match (bool) : identifier might just be contained or
                                 must match exaclty (default: True)
        """
        for i, row in enumerate(self.data):
            _ident = row[self.ident_field]
            if exact_match:
                if _ident == identifier:
                    return self.transform_func(row)
            else:
                if str(_ident).endswith(identifier):
                    return self.transform_func(row)
                elif identifier in _ident:
                    return self.transform_func(row)

    def save_record_state(self, identifier, state=None, **kwargs):
        """Mark Record state"""

        # read datasets
        if not state:
            state = self.mark['lock']
        if identifier in self.index.keys():
            (idx_raw, idx_data) = self.index[identifier]
            dict_row = self.data[idx_data]
            right_now = time.strftime(STATETIME_FORMAT)
            if kwargs:
                for k, v in kwargs.items():
                    dict_row[k] = v
            dict_row[self.state_field] = state
            dict_row[self.state_ts_field] = right_now
            self._raw_lines[idx_raw] = OAIRecordHandler._to_str_nl(dict_row)

        # if not existing_id:
        else:
            raise RuntimeError(
                'No Record for {} in {}! Cannot save state!'
                .format(identifier, self.data_path))

        # store actual state
        self._save_file()

    def _save_file(self):
        with open(self.data_path, 'w', encoding='utf-8', newline='')\
                as handle_write:
            handle_write.writelines(self._raw_lines)

    def states(self, criterias: list, set_state=RECORD_STATE_UNSET, 
        dry_run=True, verbose=False):
        """Process record states according certain criterias.

        Args:
            criterias (list): List of OAIRecordCriteria where each record
                must match all provided criterias. Defaults to a list 
                only containing OAIRecordCriteriaState(RECORD_STATE_UNSET).
            set_state (_type_, optional): Record state to set, if provided,
                and dry_run is disabled. Defaults to RECORD_STATE_UNSET.
            dry_run (bool, optional): Whether to persist possible
                modifications or just simulate. Defaults to True.
            verbose (bool, optional): Whether to list each record info instead of just 
                counting numbers. Defaults to False.
        """
        total_matches = []
        if len(criterias) == 0:
            criterias = [OAIRecordCriteriaState(RECORD_STATE_UNSET)]
        for record in self.data:
            if all(map(lambda c, d=record: c.matched(d), criterias)):
                if not dry_run:
                    record[self.state_field] = set_state
                    raw_index = self.index[record[self.ident_field]][0]
                    self._raw_lines[raw_index] = OAIRecordHandler._to_str_nl(record)
                total_matches.append(record)
        if not dry_run:
            self._save_file()
        if verbose:
            _report_stdout(total_matches)
        return len(total_matches)

    def frame(self, start, frame_size=1000, mark_state=RECORD_STATE_MASK_FRAME,
              sort_by=None) -> str:
        """
        create frame from OAI Records with start (inclusive)
        and opt frame_size (how many records from start)
        *please note*
        record count starts with "1", although intern represented as list
        """

        file_ext = None
        org_start = start
        if start > 0:
            start -= 1
        max_frame_size = len(self.data) - start
        end_frame = start + frame_size
        if frame_size > max_frame_size:
            end_frame = len(self.data)
        path_dir = os.path.abspath(os.path.dirname(self.data_path))
        file_name = os.path.basename(self.data_path)
        if '.' in file_name:
            file_name, file_ext = tuple(file_name.split('.'))
        if not file_ext:
            file_ext = 'csv'

        the_rows = []
        for i, row in enumerate(self.data):
            if i < start or i >= end_frame:
                row[F_STATE] = mark_state
            the_rows.append(row)

        # optional: sort by
        if sort_by is not None:
            if sort_by in self.header:
                the_rows = sorted(the_rows, key=lambda r: r[sort_by])
            else:
                raise RuntimeError("invalid sort by {}! only {} permitted!".format(
                    sort_by, self.header
                ))

        file_name_out = "{}_{:02d}_{:02d}.{}".format(
            file_name, org_start, end_frame, file_ext)
        path_out = os.path.join(path_dir, file_name_out)
        with open(path_out, 'w', encoding='UTF-8') as writer:
            csv_writer = csv.DictWriter(
                writer, delimiter='\t', fieldnames=self.header)
            csv_writer.writeheader()
            for row in the_rows:
                csv_writer.writerow(row)
        return path_out

    def merges(self, other_handler,
               other_require_state=None, other_ignore_state=RECORD_STATE_UNSET,
               append_unknown=True, dry_run=True, verbose=False) -> dict:
        """Merge record data into this list from other_list.
        Detected other_record with known identifier (via index) and
        in case of matching identifiers, merge *only* very first match.
        Merge only record data if source record is *not* in other_ignore_state.

        Precondition: both header fields must fit

        Args:
            other_handler (str|OAIRecordHandler):
                other OAIRecordHandler or str representation of file path
            other_require_state (str, default: None): only respect states from
                other_handler matching this state, if set. Can be used to
                compare or pick just specific outcome like 'migration_done',
                'ocr_done' or alike
            other_ignore_state (str, default: RECORD_STATE_UNSET): ignore record
                from other_handler with this state to preserve existing
                information in record from self_handler
            append_unknown (boolean, default:True): Whether append unknown
                record data or, if set 'False', leave them where they are
            dry_run (bool, default:True): Whether try/analyze or write result.
                Defaults to True, since it's a destructive operation.
            verbose (boolean, default:False): If turned on, list the records
                in detail instead of printing counts.

        Returns:
            Tuple: matched records, ignored records, new records in self from other
        """

        matches = []
        misses = []
        merges = []
        ignores = []
        requireds = []
        appendeds = []
        other_records = []
        if not isinstance(other_handler, OAIRecordHandler):
            other_handler = OAIRecordHandler(other_handler,
                                             data_fields=self.header)
        # check precondition
        if self.header != other_handler.header:
            raise OAIRecordHandlerException(
                f"Missmatch headers {self.header} != {other_handler.header}")

        other_records = other_handler.data
        # what might be integrated
        for other_record in other_records:
            other_ident = other_record[other_handler.ident_field]
            other_state = other_record[other_handler.state_field]
            # maybe merge
            if other_ident in self.index:
                idx_raw, idx_data = self.index[other_ident]
                self_record = self.data[idx_data]
                matches.append(other_record)
                if _is_unset(self_record) or _other_is_newer(self_record, other_record):
                    # preserve any existing record data
                    if other_ignore_state is not None and other_state == other_ignore_state:
                        ignores.append(other_record)
                        continue
                    if other_require_state is not None:
                        if other_state == other_require_state:
                            requireds.append(other_record)
                        else:
                            continue
                    # store
                    if not dry_run:
                        _merge(self_record, other_record)
                        merges.append(other_record)
                        self._raw_lines[idx_raw] = OAIRecordHandler._to_str_nl(self_record)
            # other_record previously unknown
            else:
                misses.append(other_record)
                # probably append at end
                if append_unknown:
                    appendeds.append(other_record)
                    if not dry_run:
                        next_idx_raw = len(self._raw_lines)
                        next_idx_data = len(self.data)
                        self._raw_lines.append(OAIRecordHandler._to_str_nl(other_record))
                        self.data.append(other_record)
                        self.index[other_ident] = (next_idx_raw, next_idx_data)
        if not dry_run:
            self._save_file()
        if verbose:
            print(f'### MATCHES ({len(matches)}) ###')
            _report_stdout(matches)
            print(f'### MERGES ({len(merges)}) ###')
            _report_stdout(merges)
            print(f'### MISSES ({len(misses)}) ###')
            _report_stdout(misses)
            print(f'### IGNORES ({len(ignores)}) ###')
            _report_stdout(ignores)
            print(f'### REQUIREDS ({len(requireds)}) ###')
            _report_stdout(requireds)
            print(f'### APPENDEDS ({len(appendeds)}) ###')
            _report_stdout(appendeds)
        return {'matches': len(matches), 'merges': len(merges),
                'misses': len(misses), 'ignores': len(ignores),
                'requireds': len(requireds), 'appendeds': len(appendeds)}


def _merge(self_record, other_record):
    self_record[F_STATE_INFO] = other_record[F_STATE_INFO]
    self_record[F_STATE] = other_record[F_STATE]
    self_record[F_STATE_TS] = other_record[F_STATE_TS]


def _is_unset(self_record):
    if self_record[F_STATE] == RECORD_STATE_UNSET:
        return True
    return False


def _other_is_newer(self_record, other_record):
    if self_record[F_STATE_TS] == RECORD_STATE_UNSET:
        return True
    else:
        dst_time = self_record[F_STATE_TS]
        src_time = other_record[F_STATE_TS]
        if src_time > dst_time:
            return True
        return False


def _report_stdout(list_records, delimiter='\t'):
    """Print set of OAIRecords, assume they are OrderedDicts
    with identical Headers"""
    if len(list_records) > 0:
        # read header
        _header = delimiter.join(list_records[0].keys())
        print('\n' + _header)
        for _r in list_records:
            print(delimiter.join(_r.values()))


class OAILoader:
    """
    Load OAI Records with corresponding metadata

    optional: post-process metadata XML after download
              to change hrefs, drop unwanted fileGroups, ...
              (default: None, i.e. no actions)
    optional: resolve linked resources like images and ocr-data
              store resources in specified directory layout
              (defaults: 'MAX' for images, 'FULLTEXT' for ocr)
              additional request arguments
              (defaults: empty dict)
    """

    def __init__(self, dir_local, base_url, **kwargs) -> None:
        self.dir_local = dir_local
        self.base_url = base_url
        self.groups = {}
        self.path_mets = None
        self.key_images = kwargs['group_images']\
            if 'group_images' in kwargs else 'MAX'
        self.post_oai = kwargs['post_oai'] if 'post_oai' in kwargs else None
        self.groups[self.key_images] = []
        self.key_ocr = kwargs['group_ocr']\
            if 'group_ocr' in kwargs else 'FULLTEXT'
        self.request_kwargs = kwargs['request_kwargs']\
            if 'request_kwargs' in kwargs else {}
        self.groups[self.key_ocr] = []
        self.store = None

    def load(self, record_identifier, local_dst, mets_digital_object_identifier=None,
             skip_resources=False, force_update=False, metadata_format='mets') -> int:
        """
        load metadata from OAI with optional caching in-between
        request additional linked resources if required

        * requires record_identifier to complete basic OAI request url
        * use mets_digital_object_identifier to get the proper MODS-section
          _if known_ for further inspection

        returns total of loaded metadata (1) plus number of
        additionally loaded resources
        """
        loaded = 0

        # sanitize url
        res_url = "{}?verb=GetRecord&metadataPrefix={}&identifier={}".format(
            self.base_url, metadata_format, record_identifier)
        self.path_mets = local_dst
        path_res = self._handle_load(res_url, self.path_mets, self.post_oai, force_update)
        if path_res:
            loaded += 1

        if skip_resources:
            return loaded

        # inspect if additional file resources are requested
        mets_reader = MetsReader(self.path_mets, mets_digital_object_identifier)

        # get linked resources
        for k in self.groups:
            self.groups[k] = mets_reader.get_filegrp_links(group=k)

        # if exist, download them too
        post_func = None
        for k, linked_res_urls in self.groups.items():
            if k == self.key_ocr:
                post_func = post_oai_store_ocr
            for linked_res_url in linked_res_urls:
                res_val_end = linked_res_url.split('/')[-1]
                res_val_path = self._calculate_path(k, res_val_end)
                if self._handle_load(linked_res_url, res_val_path, post_func):
                    loaded += 1
        return loaded

    def _handle_load(self, res_url, res_path, post_func, force_load=False):
        if self.store:
            stored_path = self.store.get(res_path)
            # if in store found ...
            if stored_path:
                if not force_load:
                    return None
                else:
                    # force update:
                    # 1. rename existing data
                    file_name = os.path.basename(stored_path)
                    file_dir = os.path.dirname(stored_path)
                    mets_ctime = str(int(os.stat(stored_path).st_mtime))
                    bkp_mets = file_name.replace('mets', mets_ctime)
                    os.rename(stored_path, os.path.join(file_dir, bkp_mets))
                    # 2. download again anyway
                    data_path = self.load_resource(res_url, res_path, post_func)
                    if data_path:
                        self.store.put(data_path)
                return res_path
            else:
                data_path = self.load_resource(res_url, res_path, post_func)
                if data_path:
                    self.store.put(data_path)
                return res_path
        else:
            return self.load_resource(res_url, res_path, post_func)

    def _calculate_path(self, *args):
        """
        calculate final path depending on some heuristics which
        fileGrp has been used - 'MAX' means images, not means 'xml'
        """
        res_path = os.path.join(str(self.dir_local), os.sep.join(list(args)))
        if '/MAX/' in res_path and not res_path.endswith('.jpg'):
            res_path += '.jpg'
        elif '/FULLTEXT/' in res_path and not res_path.endswith('.xml'):
            res_path += '.xml'
        return res_path

    def load_resource(self, url, path_local, post_func):
        """
        ensure local target dir exits and content can be written to
        optional: post-processing of received data if both exist
        """
        try:
            dst_dir = os.path.dirname(path_local)
            if not os.path.isdir(dst_dir):
                os.makedirs(dst_dir)
            local_path, data, content_type = request_resource(
                url, path_local, **self.request_kwargs)
            if post_func and data:
                # divide METS from XML (OCR-ALTO)
                # in rather brute force fashion
                _snippet = data[:200]
                # propably sanitize data, as it might originate
                # from test-data or *real* requests
                if not isinstance(_snippet, str):
                    _snippet = _snippet.decode('utf-8')
                if XMLNS['mets'] in _snippet or XMLNS['oai'] in _snippet:
                    data = post_func(self, data)
                elif 'http://www.loc.gov/standards/alto' in _snippet:
                    data = post_func(local_path, data)
                else:
                    raise OAILoadException(f"Can't handle {content_type} from {url}!")
            return local_path
        except OAILoadException as load_exc:
            raise load_exc
        except Exception as exc:
            msg = "processing '{}': {}".format(url, exc)
            raise RuntimeError(msg) from exc


class OAIFileSweeper:
    """ delete all resources (files and empty folder),
        except colorchecker, if any
        parse *.mets.xml to identify files to be deleted
    """

    def __init__(self, path_store, pattern='mets.xml', filegroups=['MAX', ]):
        self.work_dir = path_store
        self.pattern = pattern
        self.filegroups = filegroups if isinstance(filegroups, list)\
            else list(filegroups)

    def sweep(self):
        """remove OAI-Resources from given dir, if any contained"""

        work_dir = self.work_dir
        total = 0
        size = 0
        for curr_root, dirs, files in os.walk(work_dir):
            for filegroup in self.filegroups:
                if filegroup not in dirs:
                    continue

                curr_root_path = Path(curr_root)
                curr_filegroup_folder = curr_root_path / filegroup
                _files = [
                    f for f in curr_filegroup_folder.iterdir() if f.is_file()]

                if filegroup == 'MAX' and len(_files) < 2:
                    # One image only? This is likely the colorchecker! --> skip
                    # legacy
                    continue

                curr_mets_files = [
                    xml for xml in files if xml.endswith(self.pattern)]

                if curr_mets_files:
                    curr_mets_file = curr_mets_files[0]
                    curr_mets = curr_root_path.joinpath(curr_mets_file)
                    files_to_del = self._get_files(curr_mets, filegroup)

                    for pth in _files:
                        if pth.stem in files_to_del:
                            total += 1
                            size += pth.stat().st_size
                            try:
                                pth.unlink()
                                _parent = pth.parent
                                if _parent.is_dir() and\
                                   len(list(_parent.iterdir())) == 0:
                                    _parent.rmdir()
                            except PermissionError:
                                return 'cannot delete {} (insuff. permission)'\
                                    .format(pth)
        return (work_dir, total, "{} Mb".format(size >> 20))

    def _get_files(self, mets_xml, filegroup):
        xml_root = ET.parse(str(mets_xml)).getroot()
        xpath = ".//mets:fileGrp[@USE='{}']/mets:file/mets:FLocat".format(filegroup)
        locats = xml_root.findall(xpath, {"mets": "http://www.loc.gov/METS/"})
        links = [xl.get('{http://www.w3.org/1999/xlink}href') for xl in locats]
        return [Path(ln).stem for ln in links]


class LocalStore:
    """cache physical resources"""

    def __init__(self, dir_store_root, dir_local):
        self.dir_store_root = dir_store_root
        self.dir_local = dir_local

    def _calculate_path(self, path_res):
        if not isinstance(path_res, str):
            path_res = str(path_res)
        sub_path_res = path_res.replace(str(self.dir_local), '')
        if sub_path_res.startswith('/'):
            sub_path_res = sub_path_res[1:]
            if sub_path_res:
                return os.path.join(str(self.dir_store_root), sub_path_res)
        return None

    def get(self, path_res):
        """
        push resources by path into dst_dir, if
        they exist in store:
         * if single file requested, restore single file
         * if directory requested, restore from directory
        """

        path_store = self._calculate_path(path_res)
        if path_store and os.path.exists(path_store):
            dst_dir = os.path.dirname(path_res)
            if not os.path.exists(dst_dir):
                os.makedirs(dst_dir, exist_ok=True)
            if os.path.isfile(path_store):
                shutil.copy2(path_store, path_res)
            elif os.path.isdir(path_store):
                shutil.copytree(path_store, path_res)
            return path_store
        return None

    def put(self, path_res):
        """put single resource to path, for example created ocr data"""

        path_store = self._calculate_path(path_res)
        path_local_dir = os.path.dirname(path_store)
        if not os.path.exists(path_local_dir):
            os.makedirs(path_local_dir)
        # type cast to str required by python 3.5
        shutil.copy2(str(path_res), path_store)
        return path_store

    def put_all(self, src_dir, filter_ext='.xml'):
        """
        put all resources singular from last directory
        segement of current source directory to dst dir,
        if matching certain file_ext (default: "*.xml")
        """

        last_dir = os.path.basename(src_dir)
        dst_dir = os.path.join(self.dir_store_root, last_dir)
        if not os.path.exists(dst_dir):
            os.makedirs(dst_dir)
        entries = os.listdir(src_dir)
        existing = [f
                    for f in entries
                    if str(f).endswith(filter_ext)]
        stored = 0
        for ent in existing:
            src = os.path.join(src_dir, ent)
            dst = os.path.join(dst_dir, ent)
            shutil.copy(src, dst)
            stored += 1
        return stored


def request_resource(url: str, path_local: Path, **kwargs):
    """
    request resources from provided url
    * textual content is interpreted as xml and
      passed back as string for further processing
    * binary image/jpeg content is stored at path_local
      passes back path_local
    * optional params (headers, cookies, ... ) forwarded as kwargs

    Raises Exception for unknown Content-Types and requests' Errors
    (https://docs.python-requests.org/en/master/_modules/requests/exceptions/).

    Returns Tuple (response_status, result)
    """

    status = 0
    result = None
    try:
        response = requests.get(url, **kwargs)
        status = response.status_code
        if status >= 400:
            _inf = "url '{}' returned '{}'".format(url, status)
            if status < 500:
                raise OAILoadClientError(_inf)
            else:
                raise OAILoadServerError(_inf)
        if status == 200:
            content_type = response.headers['Content-Type']

            # textual xml data
            if 'text' in content_type or 'xml' in content_type:
                result = response.content
                xml_root = ET.fromstring(result)
                check_error = xml_root.find('.//error', xml_root.nsmap)
                if check_error is not None:
                    msg = "the download of {} fails due to: '{}'".format(
                        url, check_error.text)
                    raise OAILoadException(msg)
                path_local = _sanitize_local_file_extension(
                    path_local, content_type)

            # catch other content types by MIMI sub_type
            # split "<application|image>/<sub_type>"
            elif content_type.split('/')[-1] in ['image', 'jpeg', 'pdf']:
                path_local = _sanitize_local_file_extension(
                    path_local, content_type)
                if not isinstance(path_local, Path):
                    path_local = Path(path_local)
                path_local.write_bytes(response.content)

            # if we went this far, something unexpected has been returned
            else:
                msg = "download {} with unhandled content-type {}".format(
                    url, content_type)
                raise OAILoadContentException(msg)
        return (path_local, result, content_type)
    except (OSError) as exc:
        msg = "fail to download '{}' to '{}'".format(url, path_local)
        raise RuntimeError(msg) from exc


def _sanitize_local_file_extension(path_local, content_type):
    if not isinstance(path_local, str):
        path_local = str(path_local)
    if 'xml' in content_type and not path_local.endswith('.xml'):
        path_local += '.xml'
    elif 'jpeg' in content_type and not path_local.endswith('.jpg'):
        path_local += '.jpg'
    elif 'pdf' in content_type and not path_local.endswith('.pdf'):
        path_local += '.pdf'
    return path_local


def get_smtp_server():
    return smtplib.SMTP('localhost')


def send_mail(subject, message, sender, recipients):
    """Notify recipients with message from local host
    subject: str The Subject
    message: str The Message
    sender: str Email of sender
    recipients: str, list  Email(s)
    """

    if isinstance(recipients, list):
        recipients = ','.join(recipients)
    try:
        msg = MIMEText(message)
        msg['Subject'] = subject
        msg['From'] = sender
        msg['To'] = recipients + "\n"
        server = get_smtp_server()
        server.send_message(msg)
        server.quit()
        return "notification to '{}' sent: '{}'"\
               .format(recipients, message)
    except Exception as exc:
        msg = "Failed to notify '{}':'{}'\nmessage:\n'{}'\n{}!"\
            .format(recipients, subject, message, exc)
        return msg



"""Test [S]imple[A]rchive[F]ile generation format required for DSpace delivery"""

import os
import shutil
from unittest.mock import DEFAULT

import pytest

from digiflow import (
    export_data_from,
    map_contents,
    ExportMapping,
    DEFAULT_EXPORT_MAPPINGS,
    DigiFlowExportError,
    BUNDLE_PREVIEW as BP,
    BUNDLE_THUMBNAIL as BT,
)

from .conftest import (
    TEST_RES
)

SHARE_IT_EXPORT_COLLECTION = '123456789/20952'
THE_METS = 'mets.xml'
KITODO2_MONOGRAPHY_METS = '319696111.xml'
FAKE_METS = '12345678.xml'
FAKE_PDF = '12345678.pdf'


@pytest.fixture(name="monography_319696111")
def fixture_monography_319696111(tmp_path):
    """fixture for 319696111_staging_BooksizeS"""

    src_dir = os.path.join(TEST_RES, '319696111_mono')
    dst_dir = tmp_path / 'content'
    shutil.copytree(src_dir, dst_dir)
    work_dir = tmp_path / "export_workdir"
    work_dir.mkdir()
    exp_dir = tmp_path / "export_finished"
    exp_dir.mkdir()
    yield (dst_dir, work_dir, exp_dir)


def test_export_mappings_mets(monography_319696111):
    """Behavior with default paths for matchings
    between source directory (kitodo: contents)
    and working directory"""

    # arrange
    (cnt_dir, work_dir, _) = monography_319696111

    # act
    export_mappings = map_contents(cnt_dir, work_dir)

    # assert
    assert len(export_mappings) == 1
    assert export_mappings[0].path_source.endswith(KITODO2_MONOGRAPHY_METS)
    assert export_mappings[0].path_target.endswith(THE_METS)


def test_export_mappings_invalid_workdir(monography_319696111):
    """Behavior with default matching paths
    TODO clear concerns"""

    # arrange
    (w_dir, _ ,_ ) = monography_319696111

    # act
    with pytest.raises(DigiFlowExportError) as export_err:
        ExportMapping(str(w_dir)[5:], w_dir.parent)

    # assert
    assert "Invalid source path" in str(export_err.value)


EXPORT_MAPPING_KITODO2 = {'.xml': THE_METS, '.pdf': None, 'jpgs/max': None}
def test_export_kitodo2_mappings(monography_319696111):
    """Behavior with default mappings which
    correspond valid data"""

    # arrange
    (src_dir, work_dir, _) = monography_319696111

    # act
    content_files = map_contents(src_dir, work_dir, EXPORT_MAPPING_KITODO2)

    # assert
    assert len(content_files) == 69
    for content_file in content_files:
        last_src_segm = content_file.path_source.split('/')[-1]
        last_dst_segm = content_file.path_target.split('/')[-1]
        if last_dst_segm == 'mets.xml':
            continue
        assert last_src_segm == last_dst_segm


def test_export_kitodo2_result(monography_319696111):
    """
    create export with monography data from kitodo2
    """

    # arrange
    (cnt_dir, _, exp_dir) = monography_319696111
    process_metafile = os.path.join(cnt_dir, KITODO2_MONOGRAPHY_METS)

    # act
    result = export_data_from(
        process_metafile, SHARE_IT_EXPORT_COLLECTION,
        export_map=EXPORT_MAPPING_KITODO2,
        saf_final_name='319696111', export_dst=exp_dir)

    # assert
    final_location = str(os.path.join(exp_dir, '319696111.zip'))
    assert len(result) == 2
    # changed due intermediate ".processing" - suffix
    assert result[0].startswith(final_location)
    assert result[1] == '0MB'


def test_export_kitodo2_inspect_saf_assets(monography_319696111):
    """
    create export with monography data from kitodo2
    """

    # arrange
    (cnt_dir, _, exp_dir) = monography_319696111
    process_metafile = os.path.join(cnt_dir, KITODO2_MONOGRAPHY_METS)

    # act
    export_data_from(
        process_metafile, SHARE_IT_EXPORT_COLLECTION,
        export_map=EXPORT_MAPPING_KITODO2,
        saf_final_name='319696111', export_dst=exp_dir)

    # respect intermediate ".processing" - suffix
    # required for calling shutil - otherwise claims 
    # ".processing" is invalid archive format
    _tmp_saf_file_name = next(filter(lambda e: 'zip' in e, os.listdir(exp_dir)))
    _final_saf_file_name = _tmp_saf_file_name.replace('.processing', '')
    os.rename(os.path.join(exp_dir, _tmp_saf_file_name), 
              os.path.join(exp_dir, _final_saf_file_name))
    final_saf_path = str(os.path.join(exp_dir, _final_saf_file_name))
    shutil.unpack_archive(final_saf_path, extract_dir=exp_dir)

    # assert
    assert os.path.exists(os.path.join(exp_dir, 'item_000'))
    assert os.path.exists(os.path.join(exp_dir, 'item_000', '00000001.jpg'))
    assert os.path.exists(os.path.join(exp_dir, 'item_000', '319696111.pdf'))


def test_export_kitodo2_inspect_saf_contents_file(monography_319696111):
    """Inspect generated contents file"""

    # arrange
    (cnt_dir, _, exp_dir) = monography_319696111
    process_metafile = os.path.join(cnt_dir, KITODO2_MONOGRAPHY_METS)

    # act
    export_data_from(
        process_metafile, SHARE_IT_EXPORT_COLLECTION,
        export_map=EXPORT_MAPPING_KITODO2,
        saf_final_name='319696111', export_dst=exp_dir)
    
    # respect intermediate ".processing" - suffix
    # required for calling shutil - otherwise claims 
    # ".processing" is invalid archive format
    _tmp_saf_file_name = next(filter(lambda e: 'zip' in e, os.listdir(exp_dir)))
    _final_saf_file_name = _tmp_saf_file_name.replace('.processing', '')
    os.rename(os.path.join(exp_dir, _tmp_saf_file_name), 
              os.path.join(exp_dir, _final_saf_file_name))
    final_saf_path = str(os.path.join(exp_dir, _final_saf_file_name))
    shutil.unpack_archive(final_saf_path, extract_dir=exp_dir)

    # check contents file
    path_contents = os.path.join(exp_dir, 'item_000', 'contents')
    assert os.path.exists(path_contents)
    with open(path_contents, encoding='UTF-8') as cfh:
        all_lines = cfh.readlines()
        assert 69 == len(all_lines)
        contents_one_liner = ' '.join(all_lines)
        assert THE_METS in contents_one_liner
        assert "00000001.jpg\tbundle:MAX_IMAGE" in contents_one_liner
        assert "mets.xml\tbundle:FULLTEXT_OCR" not in contents_one_liner


def test_export_kitodo2_without_share_it_info_raises_exception(
        monography_319696111):
    """
    create export with monography data but invalid share_it info
    """

    # arrange
    (_, work_dir, exp_dir) = monography_319696111
    process_metafile = os.path.join(work_dir, KITODO2_MONOGRAPHY_METS)

    # act
    with pytest.raises(Exception) as exc:
        export_data_from(process_metafile, None, '319696111', 
            str(exp_dir), EXPORT_MAPPING_KITODO2)

    # assert
    assert 'No collections data provided' in str(exc.value)

    
def test_export_mappings_migration_monography(tmp_path):
    """Behavior with default matching paths
    bug fix monography 3207696
    """

    # arrange
    source_dir = os.path.join(TEST_RES, '320796')
    target_dir = tmp_path / 'WORKDIR'
    shutil.copytree(source_dir, target_dir)
    the_workdir = tmp_path / "export_workdir"
    the_workdir.mkdir()
    matching_paths = {'.xml': THE_METS, '.pdf': None, 'MAX': None}

    # act
    mapped_files = map_contents(target_dir, str(the_workdir), matching_paths)

    # assert
    assert len(mapped_files) == 17
    for mapped_file in mapped_files:
        last_src_segm = mapped_file.path_source.split('/')[-1]
        last_dst_segm = mapped_file.path_target.split('/')[-1]
        if last_dst_segm == THE_METS:
            continue
        assert last_src_segm == last_dst_segm


@pytest.fixture(name="re_ocr")
def _fixture_text_bundle(tmp_path):
    """arrange test setup for re-ocr with
    * 1 pdf
    * 1 pdf.txt
    * 10 xml files"""
    _working = tmp_path / 'WORKDIR'
    _working.mkdir()
    open(_working / FAKE_METS, mode='w').write('<mets></mets>')
    open(_working / FAKE_PDF, mode='wb').write(b'<mets></mets>')
    open(_working / f'{FAKE_PDF}.txt', mode='w').write('first line\nsecond line\nand third')
    _fulltext = _working / 'FULLTEXT'
    _fulltext.mkdir()
    _alto_txt = '<alto xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns="http://www.loc.gov/standards/alto/ns-v4#" xsi:schemaLocation="http://www.loc.gov/standards/alto/ns-v4# http://www.loc.gov/standards/alto/v4/alto-4-1.xsd"></alto>'
    for i in range(1,11):
        open(_fulltext / f'{i:08d}.xml', mode='w').write(_alto_txt)
    _export_dst = tmp_path / 'export_dir'
    _export_dst.mkdir()
    return (_working, _export_dst)


def test_export_re_ocr_with_text_bundle_mappings(re_ocr):
    """Do we have proper export mappings yet?"""

    # arrange
    (_working, _export_dst) = re_ocr
    _xport_map = DEFAULT_EXPORT_MAPPINGS
    if '.xml' in _xport_map:
        del _xport_map['.xml']
    if '.pdf' in _xport_map:
        del _xport_map['.pdf']
    _xport_map[FAKE_METS] = THE_METS
    _xport_map[FAKE_PDF] = None
    _xport_map['.pdf.txt'] = None

    # act
    mappings = map_contents(_working, _export_dst, _xport_map)

    # assert 
    assert len(mappings) == 13


@pytest.fixture(name='re_ocr_saf_item000')
def test_export_re_ocr_with_text_bundle_contents_file(re_ocr):
    """Ensure additional bundle:TEXT will be
    part of generated contents information
    """

    # arrange
    (_working, _export_dst) = re_ocr
    _xport_map = DEFAULT_EXPORT_MAPPINGS
    if '.xml' in _xport_map:
        del _xport_map['.xml']
    if '.pdf' in _xport_map:
        del _xport_map['.pdf']
    _xport_map[FAKE_METS] = THE_METS
    _xport_map[FAKE_PDF] = None
    _xport_map['.pdf.txt'] = None
    
    # act
    export_data_from(
        _working / FAKE_METS, SHARE_IT_EXPORT_COLLECTION,
        export_map=_xport_map,
        saf_final_name='12345678', export_dst=_export_dst)
    
    # respect intermediate ".processing" - suffix
    # required for calling shutil - otherwise claims 
    # ".processing" is invalid archive format
    _tmp_saf_file_name = next(filter(lambda e: 'zip' in e, os.listdir(_export_dst)))
    _final_saf_file_name = _tmp_saf_file_name.replace('.processing', '')
    os.rename(os.path.join(_export_dst, _tmp_saf_file_name), 
              os.path.join(_export_dst, _final_saf_file_name))
    final_saf_path = str(os.path.join(_export_dst, _final_saf_file_name))
    shutil.unpack_archive(final_saf_path, extract_dir=_export_dst)
    yield os.path.join(_export_dst, 'item_000')


def test_export_re_ocr_files(re_ocr_saf_item000):
    """Assert the number of files makes sense:
    * 1 mets.xml
    * 10 OCR Files
    * 1 PDF
    * 1 Text-PDF
    * 4 SAF specifics: collections, contents, dublin_core.xml, metadata_local.xml
    """

    # check contents file

    # assert
    assert os.path.exists(os.path.join(re_ocr_saf_item000, THE_METS))
    assert os.path.exists(os.path.join(re_ocr_saf_item000, FAKE_PDF))
    assert os.path.exists(os.path.join(re_ocr_saf_item000, f'{FAKE_PDF}.txt'))
    assert os.path.exists(os.path.join(re_ocr_saf_item000, 'contents'))
    assert os.path.exists(os.path.join(re_ocr_saf_item000, 'collections'))
    assert os.path.exists(os.path.join(re_ocr_saf_item000, 'dublin_core.xml'))
    assert os.path.exists(os.path.join(re_ocr_saf_item000, 'contents'))
    assert 17 == len(os.listdir(re_ocr_saf_item000))


def test_export_re_ocr_files_content_file(re_ocr_saf_item000):
    """ensure content entries present"""    
    # arrange
    path_contents = os.path.join(re_ocr_saf_item000, 'contents')
    content_entries = [l.strip() for l in open(path_contents).readlines()]

    # assert
    assert 13 == len(content_entries)
    assert 'mets.xml\tbundle:METS_BACKUP' in content_entries
    assert FAKE_PDF in content_entries
    assert f'{FAKE_PDF}.txt\tbundle:TEXT' in content_entries
    assert '00000001.xml\tbundle:FULLTEXT_OCR' in content_entries


# split two-byte marks into single byte values
JPG_MARKS = [0xff, 0xd8, 0xff, 0xd9]
@pytest.fixture(name="monography_derivates")
def _fixture_mono_derivates(tmp_path):
    """Create data with images and additional
    Share_it derivates (Defaults, Thumbnails)
    which correspond the proper DSpace
    virtual file groups"""

    _working = tmp_path / 'WORKDIR'
    _working.mkdir()
    open(_working / FAKE_METS, mode='w').write('<mets></mets>')
    open(_working / FAKE_PDF, mode='wb').write(b'<mets></mets>')
    _dirs = ['IMAGE_FOOTER', 'BUNDLE_BRANDED_PREVIEW__', 'BUNDLE_THUMBNAIL__']
    for _d in _dirs:
        _path_d = _working / _d
        _path_d.mkdir()
        for i in range(1, 11):
            _img_path = f"{_d}{i:08}.jpg"
            # clear IMAGE_FOOTER, since this is not it's name
            # on the wildside (just like '00000001.jpg')
            _img_path = _img_path.replace('IMAGE_FOOTER', '')
            open(_path_d / _img_path, 'w+b').write(bytearray(JPG_MARKS))
    _export_dst = tmp_path / 'export_dir'
    _export_dst.mkdir()
    yield (_working, _export_dst)


def test_export_migration_derivates_mappings(monography_derivates):
    """Default Mappings as from VLS migrations"""

    # arrange
    (_src, _export_dst) = monography_derivates
    _xport_map = DEFAULT_EXPORT_MAPPINGS

    # act
    mappings = map_contents(_src, _export_dst, _xport_map)

    # assert 
    assert len(mappings) == 32


@pytest.fixture(name="monography_derivates_export")
def _fixture_monography_derivates_export(monography_derivates):

    # arrange
    (_src, _export_dst) = monography_derivates
    export_data_from(
        _src / FAKE_METS, SHARE_IT_EXPORT_COLLECTION,
        saf_final_name='12345678', export_dst=_export_dst)
    
    # respect intermediate ".processing" - suffix
    # required for calling shutil - otherwise claims 
    # ".processing" is invalid archive format
    _tmp_saf_file_name = next(filter(lambda e: 'zip' in e, os.listdir(_export_dst)))
    _final_saf_file_name = _tmp_saf_file_name.replace('.processing', '')
    os.rename(os.path.join(_export_dst, _tmp_saf_file_name), 
              os.path.join(_export_dst, _final_saf_file_name))
    final_saf_path = str(os.path.join(_export_dst, _final_saf_file_name))
    shutil.unpack_archive(final_saf_path, extract_dir=_export_dst)
    yield os.path.join(_export_dst, 'item_000')


def test_export_migration_derivates_files(monography_derivates_export):
    """Ensure all generated derivates are present
    The number of the beast: 36
    * 1 mets.xml
    * 10 MAX images, 10 Thumbnails, 10 default previews
    * 1 PDF
    * 4 SAF specifics: collections, contents, dublin_core.xml, metadata_local.xml
    """

    mde = monography_derivates_export

    # assert
    assert os.path.exists(os.path.join(mde, THE_METS))
    assert os.path.exists(os.path.join(mde, FAKE_PDF))
    assert os.path.exists(os.path.join(mde, 'contents'))
    assert os.path.exists(os.path.join(mde, 'collections'))
    assert os.path.exists(os.path.join(mde, 'dublin_core.xml'))
    assert os.path.exists(os.path.join(mde, 'contents'))
    assert os.path.exists(os.path.join(mde, '00000001.jpg'))
    assert os.path.exists(os.path.join(mde, 'BUNDLE_BRANDED_PREVIEW__00000001.jpg'))
    assert os.path.exists(os.path.join(mde, 'BUNDLE_THUMBNAIL__00000001.jpg'))
    assert 36 == len(os.listdir(mde))


def test_export_migration_derivates_contents_file(monography_derivates_export):
    """The contents of the contents^TM"""

    # arrange
    path_contents = os.path.join(monography_derivates_export, 'contents')
    content_entries = [l for l in open(path_contents).readlines()]
    _img1 = '00000001.jpg'

    # assert
    assert 12 == len(content_entries)
    assert 'mets.xml\tbundle:METS_BACKUP\n' in content_entries
    assert f"{FAKE_PDF}\n" in content_entries
    assert f'{_img1}\tbundle:MAX_IMAGE\tvirtual:{BP}{_img1}/preview;{BT}{_img1}/thumbnail\n' in content_entries



# -*- coding: utf-8 -*-
"""Module for Handling of persistent Identifiers"""

import enum
import functools
import os
import re

from collections import (
    namedtuple
)
from pathlib import (
    Path
)
from typing import (
    List,
)

from .digiflow_metadata import (
    MetsReader,
    XMLNS,
)


##################
# script constants
#
# URN PATTERN
URN_PATTERN_ULB = 'urn:nbn:de:gbv:3:{}-{}-{}-{}'
DEFAULT_PATTERN_ULB = r"^urn\:nbn\:de\:gbv\:3\:\d\-\d{3,}.*\d+$"
DEFAULT_GRANULAR_PAGE_FRAGMENT = '/fragment/page'
# kitodo2 system id
KITODO2_ID = '1192015415'

###############
# Pruefziffer mapping
# Map Chars to Ints
# cf. http://www.persistent-identifier.de/?link=316
PZ_MAP = {}
PZ_MAP["0"] = 1
PZ_MAP["1"] = 2
PZ_MAP["2"] = 3
PZ_MAP["3"] = 4
PZ_MAP["4"] = 5
PZ_MAP["5"] = 6
PZ_MAP["6"] = 7
PZ_MAP["7"] = 8
PZ_MAP["8"] = 9
PZ_MAP["u"] = 11
PZ_MAP["r"] = 12
PZ_MAP["n"] = 13
PZ_MAP["b"] = 14
PZ_MAP["d"] = 15
PZ_MAP["e"] = 16
PZ_MAP[":"] = 17
PZ_MAP["a"] = 18
PZ_MAP["c"] = 19
PZ_MAP["f"] = 21
PZ_MAP["g"] = 22
PZ_MAP["h"] = 23
PZ_MAP["i"] = 24
PZ_MAP["j"] = 25
PZ_MAP["l"] = 26
PZ_MAP["L"] = 26
PZ_MAP["m"] = 27
PZ_MAP["o"] = 28
PZ_MAP["p"] = 29
PZ_MAP["q"] = 31
PZ_MAP["s"] = 32
PZ_MAP["t"] = 33
PZ_MAP["v"] = 34
PZ_MAP["w"] = 35
PZ_MAP["x"] = 36
PZ_MAP["X"] = 36
PZ_MAP["y"] = 37
PZ_MAP["Y"] = 37
PZ_MAP["z"] = 38
PZ_MAP["Z"] = 38
PZ_MAP["-"] = 39
PZ_MAP["9"] = 41
PZ_MAP["k"] = 42
PZ_MAP["_"] = 43
PZ_MAP["/"] = 45
PZ_MAP["."] = 47
PZ_MAP["+"] = 45


class DigiFlowURNExistsException(Exception):
    """Mark URN already exists"""


class GranularURNExistsException(Exception):
    """Mark Granular URN exists"""


class GranularURNException(Exception):
    """Mark generic granular URN Exception"""


class URNType(enum.IntEnum):
    '''
    Represent different level of identifiers
    Represent URN on work-level, i.e. for a
    * work-level: identifies digitalization objects like
      monographiesy, c- and f-stages or newspaper years/issues
    * page-level: physical page with complete URN
    * page-fragment-level: physical page
        with fragemented URN derived from work-level URN
    * page-fragment-focus: box within physical page
        with fragemented URN derived from work-level URN

    Please note, that there was also a granular URN
        with corresponded to only printed pages, but this
    version was not mentioned recently
    '''

    WORK = 1
    GRANULAR_PAGE = 2
    GRANULAR_FRAGMENT_PAGE = 3
    GRANULAR_FRAGMENT_PAGE_FOCUS = 4
    MISSING = 5
    UNKNOWN = 6


def generate_urn_pz(urn):
    """
    Generate Checkdigit for persistent identifiers
    see: http://www.persistent-identifier.de/?link=316
    """

    # map + reduce
    mapped_digits_str = functools.reduce(lambda c, p: c + p, [str(PZ_MAP[c]) for c in urn])

    # sum
    pzs = sum((i * int(c) for i, c in enumerate(mapped_digits_str, 1)))

    # divide by last element of digits mapped from urn
    divisor = mapped_digits_str[-1]
    quotient = int(pzs / int(divisor))

    # return last digit
    return str(quotient)[-1:][0]


# Container Struct
GranularContainer = namedtuple('GranularContainer', ['id', 'type', 'urn'])

class ProcessReport:
    """Handle outcomes concerning URN handling"""

    def __init__(self, added:List[GranularContainer], existed:List[GranularContainer]) -> None:
        self.added: List[GranularContainer] = added
        self.existed: List[GranularContainer] = existed
        self._replaced = None

    @property
    def replaced(self):
        """Get entries for which a mapping exist
        for existed and added entries, which
        means they were replaced"""
        if self._replaced is None:
            self._replaced = {}
            for _add in self.added:
                for _exs in self.existed:
                    if _add.id == _exs.id:
                        self._replaced[_add.id] = f"{_exs.urn} => {_add.urn}"
        return self._replaced


def _determine_types(elem_cnts:List) -> List[GranularContainer]:
    _types = []
    for _el in elem_cnts:
        cnt_id = _el.attrib['ID']
        gurn = _el.get('CONTENTIDS')
        if not gurn:
            _type = URNType.MISSING
        else:
            if re.match(r".*\-p\d{4}\-\d$", gurn):
                _type = URNType.GRANULAR_PAGE
            elif re.match(r".*\/fragment\/page=\d{4,}$", gurn):
                _type = URNType.GRANULAR_FRAGMENT_PAGE
            else:
                _type = URNType.UNKNOWN
        _types.append(GranularContainer(cnt_id, _type, gurn))
    return _types


def _determine_main_type(cnts:List[GranularContainer]) -> URNType:
    _map = {}
    for cnt in cnts:
        if cnt.type in _map:
            _map[cnt.type] = _map[cnt.type] + 1
        else:
            _map[cnt.type] = 1
    return max(_map.items(), key=lambda t: t[1])[0]


def enrich_urn_granular(xml_tree,
                        urn_main,
                        page_num=None,
                        padd_left=4, do_sanitize=False, do_replace=False):
    """
    Enrich URN granular at physical containers
    default: Granular URN 2.0 - GRANULAR_FRAGMENT_PAGE
    infix '/fragment/page={fptr:href[last]}' in div@CONTENTIDS
    """

    add = []
    exs = []
    cnts_total = _get_phys_containers(xml_tree)
    cnt_types = _determine_types(cnts_total)
    exs = cnt_types
    cnt_urn_fits = [e for e in cnts_total if does_fit(e, pattern=DEFAULT_PATTERN_ULB)]
    urn_type = _determine_main_type(cnt_types)
    if do_replace:
        cnt_urn_fits = []
    cnt_diffs = len(cnts_total) - len(cnt_urn_fits)
    # if no granular urns exist at all, then all must get new urn ...
    if not cnt_urn_fits:
        # identify insertion point
        phys_divs = xml_tree.findall('.//mets:structMap[@TYPE="PHYSICAL"]/mets:div/mets:div', XMLNS)
        # alert invalid data
        if len(phys_divs) == 0:
            raise GranularURNException(f"No phys pages in {xml_tree.base}!")
        # insert URNTYPE.GRANULAR_FRAGMENT_PAGE for each physical container
        for phys_div in phys_divs:
            add.append(_insert_page_fragment(phys_div, urn_main, padd_left))
        urn_type = URNType.GRANULAR_FRAGMENT_PAGE
    # if differences between phys containers exist, then only some must be
    # inserted / generated depending on existing urn type
    # might happend with additional images (colorchecker, post scans, ...)
    elif cnt_diffs:
        # here we pass physical containers, since we *might* require it
        # for sanitizing CONTENTIDS for latest added structure
        try:
            add += insert_granular_urn(urn_type, cnts_total, urn_main, page_num, padd_left)
        except GranularURNExistsException as exc:
            if do_sanitize:
                (_, _, report) = _guess_urn_granular_from_existing(xml_tree, urn_main)
                add += report.added
            else:
                raise GranularURNExistsException(exc) from exc
    _report = ProcessReport(added=add,existed=exs)
    return (xml_tree, urn_type, _report)


def _guess_urn_granular_from_existing(xml_tree, urn_main):
    """Here we run in a seldom particular failure
    driven by a missing granularen URN - let's try to fix 'em just one time!
    """
    cnts_total = _get_phys_containers(xml_tree)
    phys_divs = [e for e in cnts_total if does_fit(e, pattern=DEFAULT_PATTERN_ULB)]
    last_phys_div = cnts_total[-1]
    len_phys_div = len(cnts_total)
    if str(len_phys_div) == last_phys_div.attrib['ORDER']:
        contentids = sorted([i.attrib['CONTENTIDS'] for i in phys_divs]).pop()
        page_num = int(contentids.split('-')[-2][1:]) + 1
        xml_tree, urn_type, rep = enrich_urn_granular(xml_tree, urn_main, page_num=page_num)
        _part_rep = ProcessReport(added=rep.added, existed=[])
        return (xml_tree, urn_type, _part_rep)
    raise GranularURNException(f"Inconsistent Granular URN '{xml_tree.base}'")


def _insert_page_fragment(phys_div, main_urn: str, padd:int):
    order = int(phys_div.attrib['ORDER'])
    pf_o = f"{order:0{padd}}"
    the_gran_urn = f"{main_urn}{DEFAULT_GRANULAR_PAGE_FRAGMENT}={ pf_o}"
    phys_div.attrib['CONTENTIDS'] = the_gran_urn
    return GranularContainer(id=phys_div.attrib['ID'],
                             type=URNType.GRANULAR_FRAGMENT_PAGE,
                             urn=the_gran_urn)


def does_fit(e, pattern=DEFAULT_PATTERN_ULB):
    """
    Proper NBN URN consists of
    * prefix like urn:nbn:de:gbv:3
    * some subseqent numerical collection number
    * additional identifiers (sementic IDs, PPNs, ...)
    * suffix alike 'fragment' or '-p' for granularity
    * probably numerial exemplar count
    * final singular checkdigit
    """
    if 'CONTENTIDS' not in e.attrib:
        return False
    _id = e.attrib['CONTENTIDS']
    return re.match(pattern, _id) is not None


def insert_granular_urn(urn_type, phys_conts, main_urn, page_num=None, padd_left=4):
    """
    Take care of inconsistent physical containers!

    Physical containers missing CONTENTIDS
    http://digital.bibliothek.uni-halle.de/hd/oai/?verb=GetRecord&metadataPrefix=mets&mode=xml&identifier=10595

    Granular URN missing in between, therefore ORDER and granular URN do not match in the end
    https://digitale.bibliothek.uni-halle.de/vd16/oai?verb=GetRecord&metadataPrefix=mets&identifier=993531
    """

    inserts = []
    els_missing = [e for e in phys_conts if not does_fit(e, pattern=DEFAULT_PATTERN_ULB)]
    if urn_type == URNType.GRANULAR_PAGE:
        for el_misses in els_missing:
            # crashes if @ORDER cant be parsed as integer
            # calculate pXXXX from attr 'ORDER'
            # or, -if provided-, from parameter page_num
            p_number = page_num if page_num else int(el_misses.attrib['ORDER'])
            raw_gurn = f"{main_urn}-p{p_number:0{padd_left}}-"
            new_gurn = f"{raw_gurn}{generate_urn_pz(raw_gurn)}"
            el_id = el_misses.attrib['ID']
            # check if new granular URN is *really* unique
            # raise Exception otherwise ...
            cnts_with_gurn = [
                e for e in phys_conts
                if 'CONTENTIDS' in e.attrib and re.match(r".*-p\d+-\d$", e.attrib['CONTENTIDS'])]
            existing_ones = [e for e in cnts_with_gurn
                             if e.attrib['CONTENTIDS'] == new_gurn]
            if existing_ones:
                pat = "calculated granular URN '{}' already used by: '{}'!"
                exsting_ids = [e.attrib['ID'] for e in existing_ones]
                msg_fail = pat.format(new_gurn, exsting_ids)
                raise GranularURNExistsException(msg_fail)
            el_misses.attrib['CONTENTIDS'] = new_gurn
            inserts.append(GranularContainer(id=el_id, type=URNType.GRANULAR_FRAGMENT_PAGE, urn=new_gurn))
    return inserts


def _get_phys_containers(xml_tree):
    """Get all physical containers that contain @ORDER attribute"""
    return xml_tree.findall(
        './/mets:structMap[@TYPE="PHYSICAL"]//mets:div[@ORDER]', XMLNS)


def enrich_urn_kitodo2(process_path:'str|Path', collection='1', system_id=KITODO2_ID, exemplar='1',
               force_overwrite=False):
    """Enrich URN for Kitodo2 if not existing"""

    _urns = []
    read = MetsReader(process_path)
    ulb_urns = read.findall('.//goobi:metadata[@name="ulb_urn"]', read.primary_dmd)
    if ulb_urns and not force_overwrite:
        raise GranularURNExistsException(f"Detected existing URN {ulb_urns}!")

    identifiers = read.xpath('.//goobi:metadata[@name="CatalogIDSource" and(not(@anchorId))]/text()')
    if len(identifiers) != 1:
        raise GranularURNException(f"Ambigious/invalid base URN {identifiers}!")

    # format persistent identifier (aka URN) with provided info
    urn = URN_PATTERN_ULB.format(collection, system_id, identifiers[0], exemplar)

    # generate pruefziffer
    urn += generate_urn_pz(urn)

    # insert
    read.insert_into_prime_mods('{http://meta.goobi.org/v1.5.1/}metadata',
                    {'name': 'ulb_urn'}, urn)
    read.write()
    _urns.append(urn)

    # multi volume?
    process_path = os.path.dirname(process_path)
    path_meta_anchor = os.path.join(process_path, 'meta_anchor.xml')
    if os.path.exists(path_meta_anchor):
        ppn_c_stage = read.xpath('.//*[@anchorId="true"]/text()')
        urn_host = f'urn:nbn:de:gbv:3:{collection}-{system_id}-{ppn_c_stage[0]}-{exemplar}'
        pz_c_stage = generate_urn_pz(urn_host)
        urn_host += pz_c_stage
        _proc_host = MetsReader(path_meta_anchor)
        _proc_host.insert_into_prime_mods('{http://meta.goobi.org/v1.5.1/}metadata',
                    {'name': 'ulb_urn'}, urn_host)
        _proc_host.write()
        _urns.append(urn_host)
    return _urns



# -*- coding: utf-8 -*-

import os
import shutil

import pytest

import lxml.etree as ET

from digiflow import (
    validate_xml,
    MetsProcessor,
    MetsReader,
    XMLNS,
    MARK_KITODO2,
)

from .conftest import (
    TEST_RES,
    LIB_RES,
)

EXPORT_METS = 'export_mets.xml'


@pytest.mark.skipif("sys.version_info < (3,6)")
def test_validate_archivable_01(tmp_path):
    """Test depends on PosixPath, only works with 3.6+"""

    src_path = TEST_RES / "k2_mets_vd18_147638674.xml"
    path_meta = tmp_path / "147638674.xml"
    shutil.copyfile(src_path, path_meta)

    # act
    assert validate_xml(path_meta)


def test_create_export_mets_invalid():
    """
    xml validation of some artificial zkw export
    """

    # arrange
    path_zk_export = os.path.join(TEST_RES, 'zkw', '584', EXPORT_METS)
    assert os.path.exists(path_zk_export)

    # act
    with pytest.raises(RuntimeError) as err:
        validate_xml(path_zk_export)
    assert 'invalid schema' in err.value.args[0]
    assert 'messageDigestAlgorithm' in err.value.args[0]
    assert "value 'foo' is not an element of the set" in err.value.args[0]


def test_create_export_mets_valid():
    """
    xml validation if export_mets from kitodo2 ID 1064
    """

    # arrange
    path_zk_export = os.path.join(TEST_RES, 'zkw', '1064', EXPORT_METS)
    assert os.path.exists(path_zk_export)

    # act
    assert validate_xml(path_zk_export)


def test_altov4_is_valid():
    """Ensure validity for ALTO V4"""

    # arrange
    path_altov4_737429 = os.path.join(TEST_RES, 'ocr', 'alto', 'FULLTEXT_737438.xml')
    assert os.path.exists(path_altov4_737429)

    # act
    assert validate_xml(path_altov4_737429)


def test_altov4_from_kraken_serializer_is_valid():
    """Ensure Kraken produces valid ALTO"""

    # arrange
    path_altov4_737429 = os.path.join(
        TEST_RES, 'ocr', 'alto', 'test-kraken-alto4-serialization.xml')
    assert os.path.exists(path_altov4_737429)

    # act
    assert validate_xml(path_altov4_737429)


def test_mets_from_migration_mvwvd18_cstage_is_invalid():
    """
    Check oai:digitale.bibliothek.uni-halle.de/vd18:9427342
    when dropped filegroup DOWNLOAD - which is the only fileGroup -
    METS is invalid since fileSec *must* contain at least one fileGrp
    """

    # arrange
    path_2910519 = os.path.join(TEST_RES, 'migration', '2910519.fail.xml')
    assert os.path.exists(path_2910519)

    # act
    with pytest.raises(Exception) as exc:
        validate_xml(path_2910519)

    # assert
    _info = exc.value.args[0]
    assert 'invalid schema' in _info
    assert 'ERROR:SCHEMASV:SCHEMAV_ELEMENT_CONTENT' in _info
    assert "Element '{http://www.loc.gov/METS/}fileSec': Missing child element(s)." in _info
    assert "Expected is ( {http://www.loc.gov/METS/}fileGrp )" in _info


def test_metsreader_kitodo2_mvw():
    """
    Correct primary mods?
    """

    # arrange
    path = os.path.join(TEST_RES, 'k2_mets_vd18_183475917.xml')
    assert os.path.exists(path)
    reader = MetsReader(path)

    # act
    report = reader.analyze()

    # assert
    assert report.languages == ['ger']
    assert not report.images
    assert report.type == 'Af'
    assert 'gvk-ppn' in report.identifiers
    assert report.identifiers['gvk-ppn'] == '183475917'
    assert report.hierarchy == [('DMDLOG_0001', 'volume'), ('183475631', 'multivolume_work')]
    assert report.locations == ['Nr 83 (6)']


def test_metsreader_report_vd18_cstage():
    """
    Are correct informations extracted for c-stage?
    """

    # arrange
    path = os.path.join(TEST_RES, 'migration', '9427342.mets.xml')
    assert os.path.exists(path)
    reader = MetsReader(path, 'md9427342')

    # act
    report = reader.analyze()

    # assert
    assert report.languages == ['ger']
    assert not report.images
    assert report.type == 'Ac'
    assert report.identifiers['ulbhalvd18'] == '211999504'
    assert report.hierarchy == [('9427342', 'multivolume_work')]
    assert not report.locations


def test_metsreader_report_vd18_fstage():
    """
    Correct informations extracted for f-stage?
    """

    # arrange
    path = os.path.join(TEST_RES, 'migration', '9427337.mets.xml')
    assert os.path.exists(path)
    reader = MetsReader(path, 'md9427337')

    # act
    report = reader.analyze()

    # assert
    assert report.languages == ['ger']
    assert not report.images
    assert report.type == 'Af'
    assert report.identifiers['ulbhalvd18'] == '211999628'
    assert report.hierarchy == [
        ('9427337', 'volume'), ('9427342', 'multivolume_work')]
    assert report.locations == ['Lb 712 a (3,2)']


def test_metsreader_report_vd17_fstage_pica_case():
    """
    Correct informations extracted for VD17 F-stage?
    Is the second letter from PICA preserved?
    """

    # arrange
    path = os.path.join(TEST_RES,
                        'migration',
                        'vd17-14591176.mets.xml')
    assert os.path.exists(path)
    reader = MetsReader(path, 'md14591176')

    # act
    report = reader.analyze()

    # assert
    assert report.languages == ['ger']
    assert not report.images
    assert report.type == 'AF'
    assert report.identifiers['pon'] == '008499756'
    assert report.hierarchy == [
        ('14591176', 'volume'), ('14591136', 'multivolume_work')]
    assert report.locations == ['TM0904 (5)']


def test_metsreader_report_hd_monography():
    """
    Correct extractions for hd monography?
    """

    # arrange
    path = os.path.join(TEST_RES,
                        'vls',
                        'monography',
                        '10595.xml')
    assert os.path.exists(path)
    reader = MetsReader(path, 'md10595')

    # act
    report = reader.analyze()

    # assert
    assert report.languages == ['ger']
    assert not report.images
    assert report.type == 'Aa'
    assert report.identifiers['ulbhaldod'] == '187143188'
    assert report.hierarchy == [('10595', 'monograph')]
    assert len(report.locations) == 2
    assert report.locations == ['Pon IIg 694, FK', 'Pon IIg 689, 4° (2)']


def test_metsreader_report_kitodo2_export_monography():
    """
    Correct extractions for hd monography?
    """

    # arrange
    path = os.path.join(TEST_RES, 'k2_mets_vd18_147638674.xml')
    assert os.path.exists(path)
    reader = MetsReader(path, 'DMDLOG_0000')

    # act
    report = reader.analyze()
    # need to set this manually since
    # we do not know the kitodo ID from METS
    report.system_identifiers[MARK_KITODO2] = '1234'

    # assert
    assert report.languages == ['ger']
    assert not report.images
    assert report.type == 'Aa'
    assert report.identifiers['gvk-ppn'] == '147638674'
    assert len(report.identifiers) == 3
    assert report.system_identifiers == {'kitodo2' : '1234'}
    assert report.hierarchy == [('DMDLOG_0000', 'monograph')]
    assert report.locations == ['Pon Za 5950, QK']


def test_metsreader_logical_type_1686755_is_document():
    """
    Check digital object 1686755
    * logical type is 'document'
    * no external PICA-type present (=None)
    """

    # arrange
    mets = os.path.join(TEST_RES, 'migration/1686755.oai.xml')

    # arrange
    mets_reader = MetsReader(mets, 'md1686755')

    # assert
    outcome = mets_reader.get_type_and_hierarchy()
    assert outcome == (None, 'document', [('1686755', 'document')])


def test_metsreader_type_pica_monography():
    """This item has proper type from pica annotation"""

    # arrange
    mets = os.path.join(TEST_RES, 'migration/201517.oai.xml')

    # act
    mets_reader = MetsReader(mets, 'md201517')

    assert "Aa" in mets_reader.get_type_and_hierarchy()


def test_metsreader_wrong_logical_type():
    """
    What happens when encountered an invalid c-stage oai-response?
    Extended Test due data errors from OCR-D-Pilotproject
    => logical type monograph should be multivolume
    """

    # arrange
    mets = os.path.join(TEST_RES, 'migration/416811.mets.xml')
    mets_reader = MetsReader(mets, 'md416811')

    # has no picaType enriched
    # asserts "monograph", which is actually *wrong*
    # in this case it should be "multivolume_work"
    # f-stage 424336
    # f-stage 415691
    outcome = mets_reader.get_type_and_hierarchy()
    assert outcome == (None, 'monograph', [('416811', 'monograph')])


@pytest.fixture(name="monograph_hd_invalid_physlinks")
def _fixture_monograph_hd_10595(tmpdir):

    target_file = str(tmpdir.mkdir('vlmetadata').join('10595.mets.xml'))
    source_file = os.path.join(TEST_RES, 'migration/10595.mets.xml')
    shutil.copy(source_file, target_file)
    return str(target_file)


def test_metsreader_invalid_physlinks_10595(monograph_hd_invalid_physlinks):
    """
    Handle digital objects with invalid links from
    logical maps to physical structure
    """
    mets = MetsReader(monograph_hd_invalid_physlinks, 'md10595')
    structs = mets.get_invalid_physical_structs()

    assert len(structs) == 26
    assert structs[0] == ('log10601', 'phys2376732', 'phys10603')
    assert structs[14] == ('log2376740', 'phys2376742', 'phys10604')
    assert structs[15] == ('log2376737', 'phys2376739', 'phys10604')
    assert structs[16] == ('log2376734', 'phys2376736', 'phys10604')


def test_metsreader_report_for_10595(monograph_hd_invalid_physlinks):
    """Report for digital library object hd/10595"""

    # arrange
    reader = MetsReader(monograph_hd_invalid_physlinks, 'md10595')

    # act
    report = reader.analyze()

    # assert
    assert 'Aa' in report.type
    assert report.languages == ['ger']
    assert report.identifiers['ulbhaldod'] == '187143188'
    assert len(report.links) == 26


def test_metsreader_logical_type_is_multivolume():
    """
    Check that a multivolume_work with 4 volumes is
    correct recognized as pica 'Ac' and 'multivolume_work'
    """

    # arrange
    mets = os.path.join(TEST_RES, 'migration/9427342.mets.xml')
    mets_reader = MetsReader(mets, 'md9427342')

    # act
    mets_reader.analyze()

    # assert
    outcome = mets_reader.get_type_and_hierarchy()
    assert outcome == ('Ac', 'multivolume_work', [
                       ('9427342', 'multivolume_work')])


def test_metsreader_logical_type_is_tome():
    """
    Check that a part of a multivolume_work, 9427334,
    is VL 'tome' and pica 'Af'
    """

    # arrange
    mets = os.path.join(TEST_RES, 'migration/9427334.mets.xml')
    mets_reader = MetsReader(mets, 'md9427334')

    # act
    mets_reader.analyze()

    # assert
    outcome = mets_reader.get_type_and_hierarchy()
    assert outcome == (
        'Af', 'tome', [('9427334', 'tome'), ('9427342', 'multivolume_work')])


def test_metsreader_ambigious_recordinfo():
    """
    Handle strange cornercase in HD collection where digital objects
    posses several recordInfo elements but first match is empty
    """

    # arrange
    mets = os.path.join(TEST_RES, 'migration/369765.mets.xml')
    mets_reader = MetsReader(mets, 'md369765')

    # act
    mets_reader.analyze()

    # assert
    outcome = mets_reader.get_type_and_hierarchy()
    assert outcome == (None, 'monograph', [('369765', 'monograph')])


def test_metsreader_clear_agents(tmp_path):
    """
    Handle strange cornercase in HD collection where digital objects
    posses several recordInfo elements but th first match is empty
    """

    # arrange
    mets = os.path.join(TEST_RES, 'migration/369765.mets.xml')
    the_orig = ET.parse(mets)
    orig_agents = the_orig.findall('.//mets:agent', XMLNS)
    assert len(orig_agents) == 4
    dst = tmp_path / '369765.mets.xml'
    shutil.copyfile(mets, str(dst))
    mets_reader = MetsReader(str(dst), 'md369765')

    # act
    mets_reader.clear_agents('OTHERTYPE', ['REPOSITORY', 'INSTANCE'])
    result_path = mets_reader.write('ulb')

    # assert
    the_root = ET.parse(str(result_path))
    agents = the_root.findall('.//mets:agent', XMLNS)
    assert len(agents) == 2
    for agent in agents:
        assert 'REPOSITORY' not in agent.attrib['OTHERTYPE']
        assert 'INSTANCE' not in agent.attrib['OTHERTYPE']


def test_metsreader_log_hierarchy_menadoc_oai_record_section():
    """
    Handle strange records from menadoc with logical root element
    of type section/article which refers to a MODS section
    without any identifiers from catalogues/databases/urn
    """

    # arrange
    target_file = os.path.join(TEST_RES, 'migration/20586.mets.xml')

    # act
    _mreader = MetsReader(target_file)
    with pytest.raises(RuntimeError) as rer:
        _report = _mreader.report
    assert 'no record _identifiers' in rer.value.args[0]

    # assert - no, this is not happening!
    #assert _report.identifiers['menadoc.bibliothek.uni-halle.de/dmg'] == '20586'


def test_metsreader_opendata2_inspect_migrated_record_identifiers():
    """Inspect migrated OAI record identifiers
    from digitale.bibliothek.uni-halle.de/vd16 to
    opendata2.uni-halle.de

    Please note: for the first migrated records
    the legacy identifier format was subject to
    change, therefore the mappings must be set
    outside the METS-data
    """

    # arrange
    target_file = os.path.join(TEST_RES, 'mets/vd16_opendata2_1516514412012_4400.xml')
    mets_reader = MetsReader(target_file)

    # act
    report = mets_reader.analyze()

    # assert
    assert mets_reader.dmd_id == 'md998423'
    assert len(report.identifiers) == 5
    assert report.identifiers == {
        'urn': 'urn:nbn:de:gbv:3:1-507459',
        'vd16': 'ZV 932',
        'bvb': 'VD0034491',
        'gvk-ppn': '567526844',
        'doi': 'doi:10.25673/opendata2-4398'
    }
    assert report.system_identifiers == {
        'legacy vlid': '998423',
        'opendata2.uni-halle.de': 'https://opendata2.uni-halle.de//handle/1516514412012/4400',
    }


def test_metsreader_opendata_migrated_record_with_doi():
    """Ensure migrated data finally also contains
    a DOI identifier which was registered afterwards"""

    # arrange
    target_file = os.path.join(TEST_RES, 'opendata/1981185920_43053.xml')
    mets_reader = MetsReader(target_file)

    # act
    report = mets_reader.analyze()

    # assert
    assert mets_reader.dmd_id == 'md1177525'
    assert len(report.identifiers) == 5
    assert report.identifiers == {
        'urn': 'urn:nbn:de:gbv:3:1-132151',
        'gvk-ppn': '216311322',
        'doi': 'doi:10.25673/41099',
        'gbv': '216311322',
        'vd18': '10078320',
    }
    assert report.system_identifiers == {
        'legacy vlid': '1177525',
        'opendata.uni-halle.de': 'https://opendata.uni-halle.de//handle/1981185920/43053',
    }


def test_metsreader_opendata_inspect_migrated_record_origins():
    """How to handle migrate OAI records"""

    # arrange
    target_file = os.path.join(TEST_RES, 'opendata2/vd16-opendata2-1516514412012-4400.xml')
    mets_reader = MetsReader(target_file)

    # act
    report = mets_reader.analyze()

    # 3 origins, which is of course wrong
    assert report.origins == [('n.a.', 'Freiberg'), ('n.a.', 'Freiberg'),
                              ('digitization', 'Halle (Saale)')]


def test_metsreader_zd1_issue_16767392():
    """Check METS-Reader-output"""

    # arrange
    mets = os.path.join(TEST_RES, 'vls/zd/zd1-16767392.oai.xml')

    # act
    mets_reader = MetsReader(mets, 'md16767392')

    assert "16767392" in mets_reader.get_identifiers()['local']
    assert "issue" in mets_reader.get_type_and_hierarchy()


def test_metsreader_zd1_issue_16359609():
    """Check METS-Reader-output"""

    # arrange
    mets = os.path.join(TEST_RES, 'vls/zd/zd1-16359609.mets.xml')

    # act
    mets_reader = MetsReader(mets)

    assert "ulbhalvd:16359609" == mets_reader.get_identifiers()['local']
    assert "issue" in mets_reader.get_type_and_hierarchy()


def test_metsprocessor_clear_filegroups_migration_vd17(tmp_path):
    """Check MetsProcessor"""

    # arrange
    mets = os.path.join(TEST_RES, 'migration/vd17-14591176.mets.xml')
    the_orig = ET.parse(mets)
    orig_file_groups = the_orig.findall('.//mets:fileGrp', XMLNS)
    assert len(orig_file_groups) == 6

    dst = tmp_path / '14591176.xml'
    shutil.copyfile(mets, str(dst))
    mets_proc = MetsProcessor(str(dst), '14591176')

    # act
    mets_proc.clear_filegroups(black_list=['DEFAULT', 'THUMBS', 'MIN', 'DOWNLOAD', 'TEASER'])
    new_roots = ET.parse(str(dst))
    assert len(new_roots.findall('.//mets:fileGrp', XMLNS)) == 6


def test_metsreader_zd2_issue_18680621():
    """Check METS-Reader with new kitodo3 newspaper structure"""

    # arrange
    mets = os.path.join(TEST_RES, 'kitodo3-zd2/1021634069-18680621.xml')

    # act
    mets_reader = MetsReader(mets)

    _idents = mets_reader.get_identifiers()
    assert _idents == {
        'urn': 'urn:nbn:de:gbv:3:1-171133730-102163406918680621-11',
        'kxp-ppn': '102163406918680621',
    }
    assert mets_reader._system_identifiers() == {'kitodo3': '4583'}
    _pica, _type, _tree = mets_reader.get_type_and_hierarchy()
    assert _pica == 'AB'
    assert _type == 'issue'
    assert _tree == [('uuid-bb586c88-d9b1-3cef-9a91-4e9f4fa5cc8a', 'issue'),
                     ('n.a.', 'day'),
                     ('n.a.', 'month'),
                     ('1021634069_1868', 'year'),
                     ('1021634069', 'newspaper')]


@pytest.mark.parametrize(["mets_path", "dmd_id"], [
    (os.path.join(TEST_RES, 'migration/10595.mets.xml'), 'md10595'),
    (os.path.join(TEST_RES, 'migration/9427342.mets.xml'), 'md9427342'),
    (os.path.join(TEST_RES, 'migration/9427337.mets.xml'), 'md9427337'),
    (os.path.join(TEST_RES, 'k2_mets_vd18_058141367.xml'), 'DMDLOG_0000'),
    (os.path.join(TEST_RES, 'k2_mets_vd18_147638674.xml'), 'DMDLOG_0000'),
    (os.path.join(TEST_RES, 'vls/zd/zd1-16359609.mets.xml'), 'md16359609'),
    (os.path.join(TEST_RES, 'opendata/123456789_27949.xml'), 'md1180329'),
])
def test_metsreader_identify_prime_dmd_section(mets_path, dmd_id):

    # act
    mets_reader = MetsReader(mets_path)

    # assert
    assert mets_reader._prime_mods_id == dmd_id


def test_metsprocessor_remove_elements_and_close_tags(tmp_path):
    """Ensure: by removing elements, markup like

        <String ID="word1" CONTENT="hellu">
            </String>

        is prohibited
    """

    # arrange
    path_altov4_737429 = os.path.join(TEST_RES, 'ocr', 'alto', 'FULLTEXT_737438.xml')
    dst = tmp_path / '737438.xml'
    shutil.copyfile(path_altov4_737429, str(dst))
    mets_proc = MetsProcessor(str(dst))

    # act
    mets_proc.remove(['alto:Shape', 'alto:Processing'])
    path_result = mets_proc.write()

    # assert
    the_lines = [l.strip() for l in open(path_result, encoding='utf-8').readlines()]
    for _line in the_lines:
        assert '</String>' not in _line


def test_metsprocessor_remove_elements_no_keyerror(tmp_path):
    """Ensure legacy namespaces don't cause trouble anymore
    
    Test Target: Prevent regression
    """

    # arrange
    _a_path = os.path.join(TEST_RES, 'vls_menadoc_99454.mets.xml')
    dst = tmp_path / '99454.xml'
    shutil.copyfile(_a_path, str(dst))
    mets_proc = MetsProcessor(str(dst))
    assert len(mets_proc.xpath('//vl:sourceinfo')) == 1

    # act
    mets_proc.remove(['vl:sourceinfo'])
    path_result = mets_proc.write()

    # assert
    resl_proc = MetsProcessor(path_result)
    assert len(resl_proc.xpath('//vl:sourceinfo')) == 0


def test_metsreader_logical_hierachy_newspaper_issue():
    """
    Expect the parent hierarchy for single issue
    to be alike this:
    * newspaper 16289661
    * year 17308997
    * month 03
    * day 0322
    * issue 16359603
    """

    # arrange
    mets = os.path.join(TEST_RES, 'vls/zd/zd1-issue-16359603.zmets.xml')
    mets_reader = MetsReader(mets, 'md16359603')

    # act
    pica, log_typ, hierachy = mets_reader.get_type_and_hierarchy()

    # assert
    assert not pica
    assert log_typ == 'issue'
    assert hierachy[1:] == [('day0322', 'day'), ('month03', 'month'),
                            ('17308997', 'year'), ('16289661', 'newspaper')]


def test_metsreader_missing_struct_mapping():
    """Check behavior if METS contains logical containers
    without corresponding physical stuff - i.e. an empty
    logical section without any pages which will 
    kill Derivans 1.6.4 afterwards

    => filter em out
    """

    # arrange
    mets = os.path.join(TEST_RES, 'opendata/1981185920_43053.xml')
    mets_reader = MetsReader(mets)
    assert mets_reader.dmd_id == 'md1177525'

    # act
    with pytest.raises(RuntimeError) as _runtime_error:
        mets_reader.check()

    # assert
    assert " no link for logical section:'log1646693'" in _runtime_error.value.args[0]


DIGIFLOW_CONFIG = LIB_RES / 'digilife.ini'


def test_metsreader3_report_vd18_cstage():
    """
    Are correct informations extracted for c-stage?
    """

    # arrange
    path = os.path.join(TEST_RES, 'migration', '9427342.mets.xml')
    assert os.path.exists(path)
    reader = MetsReader(path, 'md9427342')
    reader.config = DIGIFLOW_CONFIG
    _identifier_opts = reader.config.options('identifier')
    assert len(_identifier_opts) == 14

    # act
    _report = reader.report

    # assert
    assert _report.languages == ['ger']
    assert not _report.images
    assert _report.system_identifiers == {'digitale.bibliothek.uni-halle.de/vd18': '9427342'}
    assert _report.identifiers == {'gbv': '211999504',
                                   'ulbhalvd18': '211999504',
                                   'urn': 'urn:nbn:de:gbv:3:1-636051',
                                   'vd18': '11228628'}
    assert _report.type == 'Ac'
    assert _report.hierarchy == [('9427342', 'multivolume_work')]
    assert not _report.locations


def test_metsreader3_report_vd18_fstage():
    """
    Correct informations extracted for f-stage?
    """

    # arrange
    path = os.path.join(TEST_RES, 'migration', '9427337.mets.xml')
    assert os.path.exists(path)
    reader = MetsReader(path, 'md9427337')
    reader.config = DIGIFLOW_CONFIG

    # act
    _report = reader.report

    # assert
    assert _report.languages == ['ger']
    assert not _report.images
    assert _report.type == 'Af'
    assert _report.system_identifiers == {
        'digitale.bibliothek.uni-halle.de/vd18': '9427337'  
    }
    assert _report.identifiers == {'gbv': '211999628',
                                   'ulbhalvd18': '211999628',
                                   'urn': 'urn:nbn:de:gbv:3:1-635986',
                                   'vd18': '90311817'}
    assert _report.hierarchy == [
        ('9427337', 'volume'), ('9427342', 'multivolume_work')]
    assert _report.locations == ['Lb 712 a (3,2)']


def test_metsreader_kitodo2_mena_periodical_volume_dmd_id():
    """Process rather unbalanced Kitodo2 MENA volume
    Metadata, i.e. we failed to recognize the prime
    MODS section at first sight
    """

    # arrange
    mets = os.path.join(TEST_RES, 'k2_mets_mena_12988274719564.xml')

    # act
    mets_reader = MetsReader(mets)

    # assert
    assert mets_reader.dmd_id == 'DMDLOG_0001'



"""
For the export of processes to share_it,
the generated data should be enriched and packed into a SAF.
The SAF file is a simple ZIP file,
but it is structured according to a specified structure.

Implements task #5735
"""

import os
import pathlib
import shutil
import subprocess
import sys
import tempfile

from lxml import etree as ET

from digiflow import (
    write_xml_file,
)


# default export mappings
SRC_FULLTEXT = 'FULLTEXT'
KITODO2_MAX = 'max'
SAF_METS_XML = 'mets.xml'
SRC_MAX = 'MAX'
SRC_IMAGE_FOOTER = 'IMAGE_FOOTER'
BUNDLE_PREVIEW = 'BUNDLE_BRANDED_PREVIEW__'
SRC_PREVIEW = 'BUNDLE_BRANDED_PREVIEW__'
BUNDLE_THUMBNAIL = 'BUNDLE_THUMBNAIL__'
SRC_THUMBS = 'BUNDLE_THUMBNAIL__'
DIR_IMAGES = [KITODO2_MAX, SRC_MAX, SRC_IMAGE_FOOTER]
DIR_DERIVATES = [SRC_PREVIEW, SRC_THUMBS]
DEFAULT_EXPORT_MAPPINGS = {'.xml': SAF_METS_XML,
                   '.pdf': None,
                   SRC_IMAGE_FOOTER: None,
                   SRC_THUMBS: None,
                   SRC_PREVIEW: None,
                   SRC_FULLTEXT: None
                   }
CONTENTS_FILE_MAPPINGS = {
    SRC_FULLTEXT : 'FULLTEXT_OCR',
    SRC_MAX : ''
}

class DigiFlowExportError(Exception):
    """Mark Export Exception"""


class ExportMapping:

    def __init__(self, path_source, path_target):
        if not os.path.exists(path_source):
            raise DigiFlowExportError("Invalid source path '{}'!'".format(path_source))
        if not os.path.isabs(path_source):
            raise DigiFlowExportError("Source path '{}' is not absolute!".format(path_source))
        path_target_dir = os.path.dirname(path_target)
        if not os.path.exists(path_target_dir):
            os.makedirs(path_target_dir)
        self.path_source = path_source
        self.path_target = path_target
        _path_src = pathlib.Path(path_source)
        self.src_bundle = _path_src.parent.stem

    def get(self):
        """Access mapping"""
        return (self.path_source, self.path_target)

    def copy(self):
        """copy from source to target"""
        shutil.copy(self.path_source, self.path_target)

    def __repr__(self) -> str:
        return f"{self.path_source} => {self.path_target}"

    def __eq__(self, __o: object) -> bool:
        return self.path_source == __o.path_source and self.path_target == __o.path_target
    
    def __lt__(self, __o: object) -> bool:
        self_src_base = os.path.basename(self.path_source)
        other_src_base = os.path.basename(__o.path_source)
        return self_src_base <= other_src_base


def process(export_mappings, work_dir, archive_name,
            target_collection, target_data_dir):
    """process digitalizates with images"""

    for mapping in export_mappings:
        mapping.copy()
    _handle_dublin_core_dummy(work_dir)
    _handle_dublin_core_derivates(work_dir)
    _handle_collections_file(work_dir, target_collection)
    #_handle_contents_file(work_dir)
    _handle_contents_file(work_dir, export_mappings)
    (the_tmp_path, the_filesize) = _compress(os.path.dirname(work_dir), archive_name)
    path_export_processing = _move_to_tmp_file(the_tmp_path, target_data_dir)
    return (path_export_processing, the_filesize)


def map_contents(src_dir, dst_dir, export_map=None):
    """
    Create Mappings for source dir file entries to 
    future destination dir entries.

    Serve as intermediate "SAF Building construction site".
    
    All matching entries equal or below source dir get copied 
    in flatten manner into default SAF tmp dir 'item_000'.
    If no explicite export-mapping provided, files are
    copied as-they-are, for example 
    
    Otherwise rename file accordingly. I.e., search for a
    file with extension '.xml' and turn it into 'mets.xml'
    """

    mappings = []
    if not export_map:
        export_map = {'.xml': SAF_METS_XML}

    source_paths = [os.path.join(curr, a_file)
                    for curr, _, files
                    in os.walk(src_dir)
                    for a_file in files]

    for src, dst in export_map.items():
        for source_path in source_paths:
            if src in source_path:
                # for each individual file review
                if not dst:
                    _default_dst = os.path.basename(source_path)
                    dst_path = os.path.join(dst_dir, _default_dst)
                else:
                    dst_path = os.path.join(dst_dir, dst)
                _mapping = ExportMapping(source_path, dst_path)
                # prevent re-addition
                if _mapping not in mappings:
                    mappings.append(_mapping)

    return mappings


# def _handle_contents_file(working_item_dir, only_mets=False):
def _handle_contents_file(working_item_dir, export_mappings):
    contents_file_path = os.path.join(working_item_dir, "contents")
    with open(contents_file_path, 'a', encoding='UTF-8') as contents_file:
        for mapping in sorted(export_mappings):
            _target_base = pathlib.Path(mapping.path_target)
            _name = _target_base.name
            _src_bundle = mapping.src_bundle
            if _name == SAF_METS_XML:
                contents_file.write(f"{SAF_METS_XML}\tbundle:METS_BACKUP\n")
            elif _name.endswith('.pdf'):
                contents_file.write(f"{_name}\n")
            elif _name.endswith('.pdf.txt'):
                contents_file.write(f"{_name}\tbundle:TEXT\n")
            elif _src_bundle == SRC_FULLTEXT and _is_alto(_target_base):
                contents_file.write(f"{_name}\tbundle:FULLTEXT_OCR\n")
            elif _src_bundle in DIR_IMAGES:
                image_row = _render_row(_name, working_item_dir)
                contents_file.write(image_row)
            # we know them, but we do *not* handle 'em
            # since they're already dealt at main image
            elif _src_bundle in DIR_DERIVATES:
                continue
            else:
                raise DigiFlowExportError(f"can't handle {mapping}!")


def _render_row(img_label, working_item_dir):
    """
    Render image entry in contents file
    extend with optional virtual derivates 
    for thumbnails and default images
    identified by their path segments

    Format
    1.jpg    bundle:MAX_IMAGE    virtual:BUNDLE_BRANDED_PREVIEW__1.jpg/preview;BUNDLE_THUMBNAIL__1.jpg/thumbnail
    """
    row = f"{img_label}\tbundle:MAX_IMAGE"
    all_jpgs = [_i for _i in os.listdir(working_item_dir) 
                if _i.endswith('.jpg')]
    previews = [_i for _i in all_jpgs if BUNDLE_PREVIEW in _i]
    thumbs = [_i for _i in all_jpgs if BUNDLE_THUMBNAIL in _i]
    if len(previews) > 0 and len(thumbs) > 0:
        row = f"{row}\tvirtual:{BUNDLE_PREVIEW}{img_label}/preview;{BUNDLE_THUMBNAIL}{img_label}/thumbnail"
    return row + "\n"


def _is_alto(path_file) -> bool:
    if not isinstance(path_file, str):
        path_file = str(path_file)
    if not path_file.endswith('.xml'):
        return False
    try:
        xml_tree = ET.parse(path_file).getroot()
        namespace = xml_tree.xpath('namespace-uri(.)')
        return "alto" in namespace.lower()
    except ET.ParseError as _err:
        raise DigiFlowExportError(f"Unknown OCR-Format: {_err.args}") from _err


def _compress(work_dir, archive_name):
    """
    Switched implementation since unable to de-compress zip64Format
    created with shutil.make_archive by Share_it
    """
    zip_size = -1
    zip_file_path = os.path.join(os.path.dirname(work_dir), archive_name) + '.zip'
    previous_dir = os.getcwd() 
    os.chdir(work_dir)
    cmd = 'zip -q -r {} item_000'.format(zip_file_path)
    subprocess.run(cmd, shell=True, check=True)
    os.chmod(zip_file_path, 0o666)
    zip_size = int(os.path.getsize(zip_file_path) / 1024 / 1024)
    os.chdir(previous_dir)
    return (zip_file_path, "{}MB".format(zip_size))


def _move_to_tmp_file(the_file_path, destination):
    """
    Move propably very large export data with masked name to external drive
    """
    abs_dstination = os.path.abspath(destination)
    zip_export_path = os.path.join(abs_dstination, os.path.basename(the_file_path))
    export_processing = zip_export_path + '.processing'
    if not os.path.isdir(abs_dstination):
        os.makedirs(abs_dstination, exist_ok=True)
    shutil.move(the_file_path, export_processing)
    return export_processing


def _handle_dublin_core_dummy(work_dir):
    dc_dummy_path = os.path.join(work_dir, "dublin_core.xml")
    dublin_core = ET.Element('dublin_core')
    el_title = ET.Element('dcvalue')
    el_title.set('element', 'title')
    el_title.set('qualifier', 'none')
    el_title.text = 'DUMMY'
    el_date = ET.Element('dcvalue')
    el_date.set('element', 'date')
    el_date.set('qualifier', 'issued')
    el_date.text = '1982'
    dublin_core.append(el_title)
    dublin_core.append(el_date)
    write_xml_file(dublin_core, dc_dummy_path)


def _handle_dublin_core_derivates(work_dir):
    """
    <?xml version="1.0" encoding="UTF-8"?>
    <dublin_core schema="local">
    <dcvalue element="picturegroup" qualifier="thumbnail">EXTERNAL</dcvalue>
    <dcvalue element="picturegroup" qualifier="preview">EXTERNAL</dcvalue>
    </dublin_core>
    """
    dc_path = os.path.join(work_dir, "metadata_local.xml")
    dublin_core = ET.Element('dublin_core', {'schema': 'local'})
    el_pict = ET.Element('dcvalue')
    el_pict.set('element', 'picturegroup')
    el_pict.set('qualifier', 'thumbnail')
    el_pict.text = 'EXTERNAL'
    el_prev = ET.Element('dcvalue')
    el_prev.set('element', 'picturegroup')
    el_prev.set('qualifier', 'preview')
    el_prev.text = 'EXTERNAL'
    dublin_core.append(el_pict)
    dublin_core.append(el_prev)
    write_xml_file(dublin_core, dc_path)


def _handle_collections_file(work_dir, collections):
    if collections:
        collections_path = os.path.join(work_dir, "collections")
        with open(collections_path, 'a') as collections_file:
            collections_file.write(collections)
    else:
        raise DigiFlowExportError("No collections data provided - invalid share_it export!")


def export_data_from(process_metafile_path,
                     collection,
                     saf_final_name,
                     export_dst,
                     export_map=DEFAULT_EXPORT_MAPPINGS,
                     tmp_saf_dir=None):
    """
    Main entry point to prepare, create and export specified data
    related to provided digitalization item process metadatafile_path
    """

    source_path_dir = os.path.dirname(process_metafile_path)
    tmp_dir = tempfile.gettempdir()
    prefix = 'opendata-working-'
    if tmp_saf_dir:
        tmp_dir = tmp_saf_dir
    with tempfile.TemporaryDirectory(prefix=prefix, dir=tmp_dir) as tmp_dir:
        work_dir = os.path.join(tmp_dir, saf_final_name, 'item_000')
        _mappings = map_contents(source_path_dir, work_dir, export_map)
        return process(_mappings, work_dir, saf_final_name, collection, export_dst)



"""Helper to trigger DDB-Validation for digitized prints and newspaper issues/additionals
cf. https://github.com/Deutsche-Digitale-Bibliothek/ddb-metadata-schematron-validation
"""

import os

from typing import (
    Dict,
    List,
)

from pathlib import (
    Path
)

from lxml import etree as ET

from digiflow import (
    run_command
)

from .digiflow_metadata import (
    MetsReader,
    XMLNS,
)

# trigger saxon API
SAXON_ENABLED = True
try:
    from saxonche import (
        PySaxonProcessor,
        PyXdmValue,
    )
except ModuleNotFoundError:
    SAXON_ENABLED = False

# add schematron validation language mapping
XMLNS['svrl'] = 'http://purl.oclc.org/dsdl/svrl'


# DDB Report information *not* to care about
# because we're using this for intermediate
# digitalization / migration results
DDB_IGNORE_RULES_BASIC = [
    'fileSec_02',       # fatal: no mets:fileSec[@TYPE="DEFAULT"]
    'identifier_01',    # info:  record identifier types like "bv" or "eki" not accepted
    'titleInfo_02',     # fatal: parts of work lack titel
]

# some special corner cases, when certain DDB-rules
# shall be ignored on this stage
# example:
# at migration time there's no valid href
# for the METS-pointer in the logical section, just
# the name of the SAF because Share_it creates this
# link afterwards beyond migration workflow
DDB_IGNORE_RULES_MVW = DDB_IGNORE_RULES_BASIC + [
    # fatal: metsPtr xlink:href must have valid URN
    # but at migration time this is just a reference
    # to the local host SAF
    'structMapLogical_17',
]

# periodicals (end therefore periodical_volumes)
# may have lots of extended dateIssued elements
# which seem to confuse DDB Validation
DDB_IGNORE_RULES_NEWSPAPERS = [
    'originInfo_01',
]

# included validation schema files
DDB_MEDIA_SCH = 'ddb_validierung_mets-mods-ap-digitalisierte-medien.sch'
DDB_NEWSP_SCH = 'ddb_validierung_mets-mods-ap-digitalisierte-zeitungen.sch'
DDB_MEDIA_XSL = 'ddb_validierung_mets-mods-ap-digitalisierte-medien.xsl'
DDB_NEWSP_XSL = 'ddb_validierung_mets-mods-ap-digitalisierte-zeitungen.xsl'
SCH_DIR = Path(__file__).parent / 'resources' / 'sch'
XSL_DIR = Path(__file__).parent / 'resources' / 'xsl'
PATH_MEDIA_SCH = SCH_DIR / DDB_MEDIA_SCH
PATH_NEWSP_SCH = SCH_DIR / DDB_NEWSP_SCH
PATH_MEDIA_XSL = XSL_DIR / DDB_MEDIA_XSL
PATH_NEWSP_XSL = XSL_DIR / DDB_NEWSP_XSL

# default schematron cli jar path
# resides *outside* python package
# but still on project level
CLI_DIR = Path(__file__).parent.parent / 'digiflow' / 'schematron'
PATH_CLI = str(CLI_DIR / 'schxslt-cli.jar')

# temporary report files
REPORT_FILE_SCHEMATRON = 'report_schematron.xml'
REPORT_FILE_XSLT = 'resport_xslt.xml'

# which types require special treatment?
# if these show up, switch validation logics
DIGIS_MULTIVOLUME = ['Ac', 'Af', 'AF', 'Hc', 'Hf', 'HF',
                     'volume', 'periodical', 'periodical_volume']
DIGIS_NEWSPAPER = ['issue', 'additional']

# context text len
ASSERT_DSCR_MAX_LEN = 32
ASSERT_TEXT_MAX_LEN = 96
CRITICAL_VALIDATION_ROLES = ['critical', 'error', 'fatal']
DDB_ERROR = 'ddb_error'
DDB_OTHER = 'ddb_other'


class DigiflowDDBException(Exception):
    """Mark Validation Errors"""


class FailedAssert:
    """Encapsulate individual Schematron Report
    for failed assertion"""

    def __init__(self, elem: ET.Element):
        self.id = elem.get('id')
        self.role = elem.get('role')
        self.xpath = elem.get('location')
        self._text = '. '.join(elem.xpath('svrl:text/text()', namespaces=XMLNS))
        self._description = ' '.join(elem.xpath('svrl:description/text()', namespaces=XMLNS))
        self._properties = elem.xpath('svrl:property', namespaces=XMLNS)
        self.context = ''

    def explain(self):
        """Explain the failure"""
        _msg = f'[{self.id}] '
        if len(self._properties) > 0:
            _props = '=>'.join(f"{_p.get('id')}:{_p.text}" for _p in self._properties)
            _msg = f'{_msg} {_props}'
        if len(self.context) > 0:
            _ctx = self.context
            if len(_ctx) > ASSERT_DSCR_MAX_LEN:
                _ctx = _ctx[:ASSERT_DSCR_MAX_LEN]
            _msg = f'{_msg} test:{_ctx}'
        _add_dscr = ''
        _add_text = ''
        if self._description:
            _dscr = self._description
            if len(_dscr) > ASSERT_DSCR_MAX_LEN:
                _dscr = _dscr[:ASSERT_DSCR_MAX_LEN]
            _add_dscr = _dscr + ', '
        if self._text:
            _txt = self._text.split('\n', maxsplit=1)[0]
            if len(_txt) > ASSERT_TEXT_MAX_LEN:
                _txt = _txt[:(ASSERT_TEXT_MAX_LEN-3)] + '...'
            _add_text = _txt
        if len(_add_dscr) > 2 or len(_add_text) > 0:
            _msg = f'{_msg} ({_add_dscr}{_add_text})'
        return _msg

    def add_context(self, ctx: str):
        """Add even more context on error data"""
        if ctx is not None and len(ctx) > 0:
            self.context = ctx

    def is_error(self):
        """Need special care?"""
        return self.role in CRITICAL_VALIDATION_ROLES


def ddb_validation_sch(path_mets, path_schematron=PATH_MEDIA_SCH,
                       path_schematron_bin=PATH_CLI,
                       ignore_rules=None, aggregate_errors=True) -> Dict:
    """
    Apply DDB METS/MODS Validation from specific schematron template
    to given METS/MODS data
    * due subprocess forward might raise several exceptions
    * uses schxslt-cli.jar from https://github.com/schxslt/schxslt
    """

    if not os.path.isfile(path_mets):
        raise RuntimeError(f"missing {path_mets}")
    try:
        path_tmp_output = _forward_sch_cli(path_mets, path_schematron, path_schematron_bin)
        tmp_root = ET.parse(path_tmp_output).getroot()
        reports = tmp_root.findall('{http://purl.oclc.org/dsdl/svrl}*[@role]')
        n_reports = len(reports)
        if n_reports > 0:
            # curate ignore_rules only if required
            ignore_rules = _curate_ignorances(path_mets, ignore_rules)
            rois = [(e.attrib['role'], e.attrib['id'], e.attrib['location'])
                    for e in reports
                    if e.attrib['id'] not in ignore_rules]
            return _transform_report(rois, path_mets, aggregate_errors)
        # nothing bad to report
        return {}
    except Exception as exc:
        raise RuntimeError(exc.args) from exc


def _forward_sch_cli(path_mets, path_schematron, path_schematron_bin) -> str:
    path_mets_dir = os.path.dirname(path_mets)
    path_tmp_output = os.path.join(path_mets_dir, REPORT_FILE_SCHEMATRON)
    if not os.path.isfile(path_schematron):
        raise RuntimeError(f"missing {path_schematron}")
    if not os.path.isfile(path_schematron_bin):
        raise RuntimeError(f"missing {path_schematron_bin}")
    _cmd = f"""java -jar {path_schematron_bin} -s {path_schematron}\
                -d {path_mets} -o {path_tmp_output}"""
    run_command(_cmd, 30)
    return path_tmp_output


def _curate_ignorances(path_mets, ignore_rules) -> List[str]:
    # if valid List, no need to act
    if isinstance(ignore_rules, List):
        return ignore_rules
    # ... otherwise start to curate
    _rules = []
    # if raw string with commas, turn into List!
    if isinstance(ignore_rules, str):
        _rules = ignore_rules.split(',')
    # if unset, use proper ignore rules according to digi type!
    if ignore_rules is None:
        _rules = DDB_IGNORE_RULES_BASIC
        if _matches_digi_types(path_mets, DIGIS_MULTIVOLUME):
            _rules = DDB_IGNORE_RULES_MVW
    # return empty(=turned off) ignorance otherwise
    return _rules


def ddb_validation(path_mets, digi_type='Aa', ignore_rules=None):
    """Process METS/MODS Validation which complies to Deutsche Digitale Bibliothek (DDB)
    plus applying additional set of rules which shall not be taken into account

    Further information can be found at:
        https://github.com/Deutsche-Digitale-Bibliothek/ddb-metadata-schematron-validation

    Args:
        path_mets (str|Path): resource for METS-file
        digi_type (str, optional): PICA-Type. Defaults to 'Aa'.
        ignore_rules (_type_, optional): Which Rules shall be
            ignored when evaluating validation outcome.
            Defaults to DDB_IGNORE_RULES_BASIC.
    """
    if ignore_rules is None:
        ignore_rules = DDB_IGNORE_RULES_BASIC
    _path_xslt = str(PATH_MEDIA_XSL)
    if digi_type in DIGIS_NEWSPAPER:
        _path_xslt = str(PATH_NEWSP_XSL)
    if not isinstance(path_mets, str):
        path_mets = str(path_mets)
    # just create saxon processor context once
    # several calls seem to break stuff
    with PySaxonProcessor() as proc:
        _path_tmp_result_file = _validate_and_store_result(path_mets, proc, _path_xslt)
        if not os.path.isfile(_path_tmp_result_file):
            raise DigiflowDDBException(f"missing tmp output file {_path_tmp_result_file}")
        _failures = _get_failures(path_mets, proc, _path_tmp_result_file, ignore_rules)
        _aggregated = _aggregate_failures(_failures)
        if DDB_ERROR in _aggregated:
            raise DigiflowDDBException(_aggregated[DDB_ERROR])
        os.unlink(_path_tmp_result_file)
        return _aggregated


def _aggregate_failures(fails: List[FailedAssert]) -> Dict:
    _dict = {}
    for _f in fails:
        _xpln = _f.explain()
        _role = DDB_ERROR if _f.is_error() else DDB_OTHER
        if _role in _dict:
            _prev = _dict[_role]
            _prev.append(_xpln)
            _dict[_role] = _prev
        else:
            _dict.setdefault(_role, [_xpln])
    return _dict


def _matches_digi_types(path_mets, some_types):
    """determine if any of digi_types is of some_types
    by considering both PICA and logical types"""

    _types = list(MetsReader(path_mets).get_type_and_hierarchy()[:2])
    return any(t in some_types for t in _types)


def _validate_and_store_result(path_mets, proc, path_template):
    _path_tmp = None
    try:
        xsltproc = proc.new_xslt30_processor()
        _exec = xsltproc.compile_stylesheet(stylesheet_file=path_template)
        _doc_dir = os.path.dirname(path_mets)
        _path_tmp = os.path.join(_doc_dir, REPORT_FILE_XSLT)
        _doc_dir = os.path.dirname(path_mets)
        _path_tmp = os.path.join(_doc_dir, REPORT_FILE_XSLT)
        _exec.transform_to_file(source_file=path_mets,
                                stylesheet_file=path_template,
                                output_file=_path_tmp)
    except Exception as _exc:
        raise RuntimeError(_exc) from _exc
    return _path_tmp


def _get_failures(path_mets, proc, path_report, ignore_rules) -> List[FailedAssert]:
    """Inspect results from tmp report file
    if any irregularities detected, apply XSLT2.0
    expressions from report file to gather details"""

    _failures = []
    tmp_root = ET.parse(path_report).getroot()
    reports = tmp_root.findall('svrl:*[@role]', XMLNS)
    if len(reports) > 0:
        _fails = [FailedAssert(e)
                  for e in reports
                  if e.attrib['id'] not in ignore_rules]
        try:
            _mets_doc = proc.parse_xml(xml_file_name=path_mets)
            _xp_proc = proc.new_xpath_processor()
            _xp_proc.set_context(xdm_item=_mets_doc)
            for _fail in _fails:
                _info = _xp_proc.evaluate(_fail.xpath)
                if isinstance(_info, PyXdmValue) and _info.size > 0:
                    _items = [_info.item_at(i) for i in range(0, _info.size)]
                    _ctx = ','.join(_i.string_value.strip() for _i in _items)
                    _fail.add_context(_ctx)
                _failures.append(_fail)
        except Exception as _exc:
            raise RuntimeError(_exc) from _exc
    return _failures


def _transform_report(entries, path_mets, aggregate_errors):
    """Generate dictionary from schematron report data

    Args:
        entries (list): List of Schematron report entries
        path_mets (str): path to original METS/MODS

    Returns:
        dict: mapping role => list of entries for role
    """

    rois_map = {}
    mets_root = ET.parse(path_mets).getroot()
    for entry in entries:
        _xpath = _ddb_report_location_2_lxml_xpath(entry[2])
        affecteds = mets_root.xpath(_xpath, namespaces=XMLNS)
        _hits = _transform_affecteds(affecteds)
        hit_str = ','.join(_hits)
        if entry[0] in rois_map:
            rois_map[entry[0]].append((entry[1], hit_str))
        else:
            rois_map[entry[0]] = [(entry[1], hit_str)]
    # optional aggregate 'errors' and 'fatals'
    # merge category 'fatal' into 'error'
    if aggregate_errors:
        fatals = rois_map['fatal'] if 'fatal' in rois_map else []
        errors = rois_map['error'] if 'error' in rois_map else []
        total_errors = fatals + errors
        if total_errors and len(fatals) > 0:
            rois_map['error'] = total_errors
            del rois_map['fatal']
    return rois_map


def _transform_affecteds(affecteds):
    hits = []
    for aff in affecteds:
        if 'ID' in aff.attrib:
            hits.append(aff.attrib['ID'])
        elif 'type' in aff.attrib or 'source' in aff.attrib:
            attm = ','.join([(v + "=" + k)
                             for v, k in aff.attrib.items()])
            hits.append(attm)
        else:
            # get textual content only for *real*
            # subtrees not complete prime mods-tree
            # i.e. from mods:place
            _local_tagname = ET.QName(aff).localname
            _txt = ''
            if _local_tagname != 'mods':
                _txt = '='.join([t.strip() for t in aff.itertext()])
            hits.append(f"{__strip_namespaces(aff.tag)}{_txt}")
    return hits


def _ddb_report_location_2_lxml_xpath(xp_in):
    """Transform schematron output from DDB Report location data

    Args:
        xp_in (str): schematron failed-assert|success location attribute

    Returns:
        str: XPath stripped any qualified namespace annotations + leading '/'
    """

    return '/' + __strip_namespaces(xp_in)


def __strip_namespaces(elem_str):
    for _k, _v in XMLNS.items():
        if _v in elem_str:
            elem_str = elem_str.replace(_v, _k + ':')
    return elem_str.translate(str.maketrans({'Q': '', '{': '', '}': ''}))
